# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections import defaultdict
from functools import partial
import re
from pathlib import Path
from subprocess import CalledProcessError

import pytest

from loki.logging import WARNING
from loki import Array, Dimension, FindNodes, Literal, Sourcefile
from loki.batch import Pipeline, ProcessingStrategy, Scheduler, SchedulerConfig, Transformation
from loki.frontend import HAVE_FP, HAVE_OMNI, OMNI, available_frontends
from loki.ir import nodes as ir
from loki.transformations import (
    CMakePlanTransformation, DependencyTransformation, FileWriteTransformation,
    ModuleWrapTransformation
)

pytestmark = pytest.mark.skipif(not HAVE_FP, reason='Fparser not available')


def test_scheduler_empty_config(testdir, frontend, tmp_path):
    """
    Test that instantiating the Scheduler without config works (albeit it's not very useful)
    This fixes #373
    """
    projA = testdir/'sources/projA'

    scheduler = Scheduler(
        paths=projA, includes=projA/'include',
        seed_routines=['driverA'], frontend=frontend, xmods=[tmp_path]
    )
    assert scheduler.items == ('drivera_mod#drivera',)


def test_scheduler_cmake_planner(tmp_path, testdir, frontend):
    """
    Test the plan generation feature over a call hierarchy spanning two
    distinctive projects.

    projA: driverB -> kernelB -> compute_l1<replicated> -> compute_l2
                         |
    projB:          ext_driver -> ext_kernel
    """

    sourcedir = testdir/'sources'
    proj_a = sourcedir/'projA'
    proj_b = sourcedir/'projB'

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel',
            'expand': True,
            'strict': True,
            'ignore': ('header_mod',),
            'mode': 'foobar'
        },
        'routines': {
            'driverB': {'role': 'driver'},
            'kernelB': {'ignore': ['ext_driver']},
        }
    })
    builddir = tmp_path/'scheduler_cmake_planner_dummy_dir'
    builddir.mkdir(exist_ok=True)

    scheduler = Scheduler(
        paths=[proj_a, proj_b], includes=proj_a/'include',
        config=config, frontend=frontend, xmods=[tmp_path],
        output_dir=builddir
    )

    planfile = builddir/'loki_plan.cmake'

    # Populate the scheduler
    # (this is the same as SchedulerA in test_scheduler_dependencies_ignore, so no need to
    # check scheduler set-up itself)
    # Apply the transformation
    scheduler.process(FileWriteTransformation(), proc_strategy=ProcessingStrategy.PLAN)
    scheduler.write_cmake_plan(filepath=planfile, rootpath=sourcedir)

    plan_pattern = re.compile(r'set\(\s*(\w+)\s*(.*?)\s*\)', re.DOTALL)

    loki_plan = planfile.read_text()
    plan_dict = {k: v.split() for k, v in plan_pattern.findall(loki_plan)}
    plan_dict = {k: {Path(s).stem for s in v} for k, v in plan_dict.items()}

    # Validate the plan file content
    expected_files = {
        'driverB_mod', 'kernelB_mod',
        'compute_l1_mod', 'compute_l2_mod'
    }

    assert 'LOKI_SOURCES_TO_TRANSFORM' in plan_dict
    assert plan_dict['LOKI_SOURCES_TO_TRANSFORM'] == expected_files

    assert 'LOKI_SOURCES_TO_REMOVE' in plan_dict
    assert plan_dict['LOKI_SOURCES_TO_REMOVE'] == expected_files

    assert 'LOKI_SOURCES_TO_APPEND' in plan_dict
    assert plan_dict['LOKI_SOURCES_TO_APPEND'] == {
        f'{name}.foobar' for name in expected_files
    }

    planfile.unlink()
    builddir.rmdir()


@pytest.mark.parametrize('prec', ['DP', 'SP'])
def test_scheduler_cmake_planner_libs(tmp_path, testdir, frontend, prec):
    """
    Test the plan generation feature over a call hierarchy spanning two
    distinctive projects. However, this time using the 'lib' attribute.

    projA: driverB -> kernelB -> compute_l1<replicated> -> compute_l2
                         |
    projB:          ext_driver -> ext_kernel
    """

    sourcedir = testdir/'sources'
    proj_a = sourcedir/'projA'
    proj_b = sourcedir/'projB'

    config = SchedulerConfig.from_dict({
        'default': {
            'role': 'kernel',
            'expand': True,
            'strict': True,
            'ignore': ('header_mod',),
            'mode': 'foobar',
            'lib': f'projAlib.{prec}'
        },
        'routines': {
            'driverB': {'role': 'driver'},
            'ext_driver': {'lib': f'projBlib.{prec}'}
        }
    })
    builddir = tmp_path/'scheduler_cmake_planner_libs_dummy_dir'
    builddir.mkdir(exist_ok=True)

    scheduler = Scheduler(
        paths=[proj_a, proj_b], includes=proj_a/'include',
        config=config, frontend=frontend, xmods=[tmp_path],
        output_dir=builddir
    )

    planfile = builddir/'loki_plan_libs.cmake'
    # Populate the scheduler
    # Apply the transformation
    scheduler.process(FileWriteTransformation(), proc_strategy=ProcessingStrategy.PLAN)
    scheduler.write_cmake_plan(filepath=planfile, rootpath=sourcedir)

    loki_plan = planfile.read_text()
    plan_pattern = re.compile(r'set\(\s*(\w+)\s*(.*?)\s*\)', re.DOTALL)
    plan_dict = {k: v.split() for k, v in plan_pattern.findall(loki_plan)}
    plan_dict = {k: {Path(s).stem for s in v} for k, v in plan_dict.items()}

    # Validate the plan file content
    expected_keys = {
        'LOKI_SOURCES_TO_TRANSFORM', 'LOKI_SOURCES_TO_APPEND', 'LOKI_SOURCES_TO_REMOVE',
        f'LOKI_SOURCES_TO_TRANSFORM_projBlib_{prec}', f'LOKI_SOURCES_TO_APPEND_projBlib_{prec}',
        f'LOKI_SOURCES_TO_REMOVE_projBlib_{prec}', f'LOKI_SOURCES_TO_TRANSFORM_projAlib_{prec}',
        f'LOKI_SOURCES_TO_APPEND_projAlib_{prec}', f'LOKI_SOURCES_TO_REMOVE_projAlib_{prec}'
    }

    assert set(plan_dict.keys()) == expected_keys

    expected_files_a = {
        'driverB_mod', 'kernelB_mod',
        'compute_l1_mod', 'compute_l2_mod',
    }
    expected_files_b = {'ext_driver_mod', 'ext_kernel'}
    expected_files = expected_files_a | expected_files_b

    assert plan_dict['LOKI_SOURCES_TO_TRANSFORM'] == expected_files
    assert plan_dict['LOKI_SOURCES_TO_REMOVE'] == expected_files
    assert plan_dict['LOKI_SOURCES_TO_APPEND'] == {
        f'{name}.foobar' for name in expected_files
    }

    assert plan_dict[f'LOKI_SOURCES_TO_TRANSFORM_projAlib_{prec}'] == expected_files_a
    assert plan_dict[f'LOKI_SOURCES_TO_REMOVE_projAlib_{prec}'] == expected_files_a
    assert plan_dict[f'LOKI_SOURCES_TO_APPEND_projAlib_{prec}'] == {
        f'{name}.foobar' for name in expected_files_a
    }

    assert plan_dict[f'LOKI_SOURCES_TO_TRANSFORM_projBlib_{prec}'] == expected_files_b
    assert plan_dict[f'LOKI_SOURCES_TO_REMOVE_projBlib_{prec}'] == expected_files_b
    assert plan_dict[f'LOKI_SOURCES_TO_APPEND_projBlib_{prec}'] == {
        f'{name}.foobar' for name in expected_files_b
    }

    planfile.unlink()
    builddir.rmdir()


def test_scheduler_disable_wildcard(testdir, config, tmp_path):
    fcode_mod = """
module field_mod
  type field2d
    contains
    procedure :: init => field_init
  end type

  type field3d
    contains
    procedure :: init => field_init
  end type

  contains
    subroutine field_init()

    end subroutine
end module
"""

    fcode_driver = """
subroutine my_driver
  use field_mod, only: field2d, field3d, field_init
implicit none

  type(field2d) :: a, b
  type(field3d) :: c, d

  call a%init()
  call b%init()
  call c%init()
  call field_init(d)
end subroutine my_driver
"""

    # Set up the test files
    dirname = testdir/'test_scheduler_disable_wildcard'
    dirname.mkdir(exist_ok=True)
    modfile = dirname/'field_mod.F90'
    modfile.write_text(fcode_mod)
    testfile = dirname/'test.F90'
    testfile.write_text(fcode_driver)

    config['default']['disable'] = ['*%init']

    scheduler = Scheduler(paths=dirname, seed_routines=['my_driver'], config=config, xmods=[tmp_path])

    expected_items = ['#my_driver', 'field_mod#field_init']
    expected_dependencies = [('#my_driver', 'field_mod#field_init')]

    assert all(n in scheduler.items for n in expected_items)
    assert all(e in scheduler.dependencies for e in expected_dependencies)

    assert 'field_mod#field2d%init' not in scheduler.items
    assert 'field_mod#field3d%init' not in scheduler.items

    try:
        # Clean up
        modfile.unlink()
        testfile.unlink()
        dirname.rmdir()
    except FileNotFoundError:
        pass


def test_transformation_config(config):
    """
    Test the correct instantiation of :any:`Transformation` objecst from config
    """
    my_config = config.copy()
    my_config['transformations'] = {
        'DependencyTransformation': {
            'module': 'loki.transformations.build_system',
            'options': {
                'suffix': '_rick',
                'module_suffix': '_roll',
                'replace_ignore_items': False,
            }
        },
        'IdemTransformation': {
            'module': 'loki.transformations',
        }
    }
    cfg = SchedulerConfig.from_dict(my_config)
    assert cfg.transformations['DependencyTransformation']

    # Instantiate IdemTransformation entry without options
    transformation = cfg.transformations['DependencyTransformation']
    assert isinstance(transformation, DependencyTransformation)
    assert transformation.suffix == '_rick'
    assert transformation.module_suffix == '_roll'
    assert not transformation.replace_ignore_items

    bad_config = config.copy()
    bad_config['transformations'] = {
        'DependencyTrafo': {
            'module': 'loki.transformations.build_system',
            'options': {}
        }
    }
    # Test for errors when failing to instantiate a transformation
    with pytest.raises(RuntimeError):
        SchedulerConfig.from_dict(bad_config)

    worse_config = config.copy()
    worse_config['transformations'] = {
        'DependencyTransform': {
            # <= typo
            'module': 'loki.transformats.build_system',
            'options': {}
        }
    }
    with pytest.raises(ModuleNotFoundError):
        SchedulerConfig.from_dict(worse_config)

    worst_config = config.copy()
    worst_config['transformations'] = {
        'DependencyTransform': {
            'module': 'loki.transformations.build_system',
            # <= typo
            'options': {'hello': 'Dave'}
        }
    }
    with pytest.raises(RuntimeError):
        SchedulerConfig.from_dict(worst_config)


def test_transformation_config_external_with_dimension(testdir, config):
    """
    Test instantiation of :any:`Transformation` from config with
    :any:`Dimension` argument.
    """
    my_config = config.copy()
    my_config['dimensions'] = {
        'ij': {'size': 'n', 'index': 'i'}
    }
    my_config['transformations'] = {
        'CallMeRick': {
            'classname': 'CallMeMaybeTrafo',
            'module': 'call_me_trafo',
            'path': str(testdir/'sources'),
            'options': {'name': 'Rick', 'horizontal': '%dimensions.ij%'}
        }
    }
    cfg = SchedulerConfig.from_dict(my_config)
    assert cfg.transformations['CallMeRick']

    transformation = cfg.transformations['CallMeRick']
    # We don't have the type, so simply check the class name
    assert type(transformation).__name__ == 'CallMeMaybeTrafo'
    assert transformation.name == 'Rick'
    assert isinstance(transformation.horizontal, Dimension)
    assert transformation.horizontal.size == 'n'
    assert transformation.horizontal.index == 'i'


@pytest.mark.parametrize('item_name,keys,use_pattern_matching,match_item_parents,expected', [
    ('comp2', 'comp2', True, True, ('comp2',)),
    ('#comp2', 'comp2', True, True, ('comp2',)),
    ('comp2', '#comp2', True, True, ()),
    ('#comp2', '#comp2', True, True, ('#comp2',))
])
def test_scheduler_config_match_item_keys(item_name, keys, use_pattern_matching, match_item_parents, expected):
    # This is key: If the config key is provided with explicit scope,
    # we don't match unscoped names
    value = SchedulerConfig.match_item_keys(item_name, keys, use_pattern_matching, match_item_parents)
    assert value == expected


@pytest.mark.parametrize('frontend', available_frontends())
@pytest.mark.parametrize('frontend_args,defines,preprocess,has_cpp_directives,additional_dependencies', [
    (
        None, None, False,
        ['#test_scheduler_frontend_args1', '#test_scheduler_frontend_args2', '#test_scheduler_frontend_args4'], {
        '#test_scheduler_frontend_args2': ('#test_scheduler_frontend_args3',),
        '#test_scheduler_frontend_args3': (),
        '#test_scheduler_frontend_args4': ('#test_scheduler_frontend_args3',),
        }
    ),
    (None, ['SOME_DEFINITION'], True, [], {}),
    ({'file3_4.F90': {'defines': ['SOME_DEFINITION', 'LOCAL_DEFINITION']}}, ['SOME_DEFINITION'], True, [], {
        '#test_scheduler_frontend_args3': (),
        '#test_scheduler_frontend_args4': ('#test_scheduler_frontend_args3',),
    }),
    ({'file2.F90': {'preprocess': False}}, ['SOME_DEFINITION'], True, ['#test_scheduler_frontend_args2'], {
        '#test_scheduler_frontend_args2': ('#test_scheduler_frontend_args3',),
        '#test_scheduler_frontend_args3': (),
    }),
    ({'file2.F90': {'preprocess': True, 'defines': ['SOME_DEFINITION']}}, None, False,
     ['#test_scheduler_frontend_args1', '#test_scheduler_frontend_args4'], {
         '#test_scheduler_frontend_args3': (),
         '#test_scheduler_frontend_args4': ('#test_scheduler_frontend_args3',),
     }),
])
def test_scheduler_frontend_args(tmp_path, frontend, frontend_args, defines, preprocess,
                                 has_cpp_directives, additional_dependencies, config):
    """
    Test overwriting frontend options via Scheduler config
    """

    fcode1 = """
subroutine test_scheduler_frontend_args1
    implicit none
#ifdef SOME_DEFINITION
    call test_scheduler_frontend_args2
#endif
end subroutine test_scheduler_frontend_args1
    """.strip()

    fcode2 = """
subroutine test_scheduler_frontend_args2
    implicit none
#ifndef SOME_DEFINITION
    call test_scheduler_frontend_args3
#endif
    call test_scheduler_frontend_args4
end subroutine test_scheduler_frontend_args2
    """.strip()

    fcode3_4 = """
subroutine test_scheduler_frontend_args3
implicit none
end subroutine test_scheduler_frontend_args3

subroutine test_scheduler_frontend_args4
implicit none
#ifdef LOCAL_DEFINITION
    call test_scheduler_frontend_args3
#endif
end subroutine test_scheduler_frontend_args4
    """.strip()

    (tmp_path/'file1.F90').write_text(fcode1)
    (tmp_path/'file2.F90').write_text(fcode2)
    (tmp_path/'file3_4.F90').write_text(fcode3_4)

    expected_dependencies = {
        '#test_scheduler_frontend_args1': ('#test_scheduler_frontend_args2',),
        '#test_scheduler_frontend_args2': ('#test_scheduler_frontend_args4',),
        '#test_scheduler_frontend_args4': (),
    }

    # No preprocessing, thus all call dependencies are included
    # Global preprocessing setting SOME_DEFINITION, removing dependency on 3
    # Global preprocessing with local definition for one file, re-adding a dependency on 3
    # Global preprocessing with preprocessing switched off for 2
    # No preprocessing except for 2
    for key, value in additional_dependencies.items():
        expected_dependencies[key] = expected_dependencies.get(key, ()) + value

    config['frontend_args'] = frontend_args

    scheduler = Scheduler(
        paths=[tmp_path], config=config, seed_routines=['test_scheduler_frontend_args1'],
        frontend=frontend, defines=defines, preprocess=preprocess, xmods=[tmp_path]
    )

    assert set(scheduler.items) == set(expected_dependencies)
    assert set(scheduler.dependencies) == {
        (a, b) for a, deps in expected_dependencies.items() for b in deps
    }

    for item in scheduler.items:
        cpp_directives = FindNodes(ir.PreprocessorDirective).visit(item.ir.ir)
        # NB: OMNI always does preprocessing, therefore we won't find the CPP directives
        #     after the full parse
        assert bool(cpp_directives) == (item in has_cpp_directives and frontend != OMNI)


@pytest.mark.skipif(not (HAVE_OMNI and HAVE_FP), reason='OMNI or FP not available')
def test_scheduler_frontend_overwrite(tmp_path, config):
    """
    Test the use of a different frontend via Scheduler config
    """
    fcode_header = """
module test_scheduler_frontend_overwrite_header
    implicit none
    type some_type
        ! We have a comment
        real, dimension(:,:), pointer :: arr
    end type some_type
end module test_scheduler_frontend_overwrite_header
    """.strip()
    fcode_kernel = """
subroutine test_scheduler_frontend_overwrite_kernel
    use test_scheduler_frontend_overwrite_header, only: some_type
    implicit none
    type(some_type) :: var
end subroutine test_scheduler_frontend_overwrite_kernel
    """.strip()

    (tmp_path/'test_scheduler_frontend_overwrite_header.F90').write_text(fcode_header)
    (tmp_path/'test_scheduler_frontend_overwrite_kernel.F90').write_text(fcode_kernel)

    # Make sure that OMNI cannot parse the header file
    with pytest.raises(CalledProcessError):
        Sourcefile.from_source(fcode_header, frontend=OMNI, xmods=[tmp_path])

    # ...and that the problem exists also during Scheduler traversal
    with pytest.raises(CalledProcessError):
        Scheduler(
            paths=[tmp_path], config=config, seed_routines=['test_scheduler_frontend_overwrite_kernel'],
            frontend=OMNI, xmods=[tmp_path]
        )

    fcode_header_lines = fcode_header.split('\n')
    # Strip the comment from the header file and parse again to generate an xmod
    Sourcefile.from_source('\n'.join(fcode_header_lines[:3] + fcode_header_lines[4:]), frontend=OMNI, xmods=[tmp_path])

    # Setup the config with the frontend overwrite
    config['frontend_args'] = {
        'test_scheduler_frontend_overwrite_header.F90': {'frontend': 'FP'}
    }

    scheduler = Scheduler(
        paths=[tmp_path], config=config, seed_routines=['test_scheduler_frontend_overwrite_kernel'],
        frontend=OMNI, xmods=[tmp_path]
    )

    # ...and now it works fine
    assert set(scheduler.items) == {
        '#test_scheduler_frontend_overwrite_kernel', 'test_scheduler_frontend_overwrite_header#some_type'
    }

    assert set(scheduler.dependencies) == {
        ('#test_scheduler_frontend_overwrite_kernel', 'test_scheduler_frontend_overwrite_header#some_type')
    }

    comments = FindNodes(ir.Comment).visit(scheduler['test_scheduler_frontend_overwrite_header#some_type'].ir.body)
    assert len(comments) == 1
    # ...and the derived type has it's comment
    assert comments[0].text == '! We have a comment'


def test_scheduler_pipeline_simple(testdir, config, frontend, tmp_path):
    """
    Test processing a :any:`Pipeline` over a simple call-tree.

    projA: driverA -> kernelA -> compute_l1 -> compute_l2
                           |
                           | --> another_l1 -> another_l2
    """
    projA = testdir/'sources/projA'

    scheduler = Scheduler(
        paths=projA, includes=projA/'include', config=config,
        seed_routines='driverA', frontend=frontend, xmods=[tmp_path]
    )

    class ZeroMyStuffTrafo(Transformation):
        """ Fill each argument array with 0.0 """

        def transform_subroutine(self, routine, **kwargs):
            for v in routine.variables:
                if isinstance(v, Array):
                    routine.body.append(ir.Assignment(lhs=v, rhs=Literal(0.0)))

    class AddSnarkTrafo(Transformation):
        """ Add a snarky comment to the zeroing """

        def __init__(self, name='Rick'):
            self.name = name

        def transform_subroutine(self, routine, **kwargs):
            # Add a newline
            routine.body.append(ir.Comment(text=''))
            routine.body.append(ir.Comment(text=f'! Sorry {self.name}, no values for you!'))

    def has_correct_assigns(routine, num_assign, values=None):
        assigns = FindNodes(ir.Assignment).visit(routine.body)
        values = values or [0.0]
        return len(assigns) == num_assign and all(a.rhs in values for a in assigns)

    def has_correct_comments(routine, name='Dave'):
        text = f'! Sorry {name}, no values for you!'
        comments = FindNodes(ir.Comment).visit(routine.body)
        return len(comments) > 2 and comments[-1].text == text

    # First apply in sequence and check effect
    scheduler.process(transformation=ZeroMyStuffTrafo())
    assert has_correct_assigns(scheduler['drivera_mod#drivera'].ir, 0)
    assert has_correct_assigns(scheduler['kernela_mod#kernela'].ir, 2)
    assert has_correct_assigns(scheduler['compute_l1_mod#compute_l1'].ir, 1)
    assert has_correct_assigns(scheduler['compute_l2_mod#compute_l2'].ir, 2, values=[66.0, 00])
    assert has_correct_assigns(scheduler['#another_l1'].ir, 1)
    assert has_correct_assigns(scheduler['#another_l2'].ir, 2, values=[77.0, 00])

    scheduler.process(transformation=AddSnarkTrafo(name='Dave'))
    assert has_correct_comments(scheduler['drivera_mod#drivera'].ir)
    assert has_correct_comments(scheduler['kernela_mod#kernela'].ir)
    assert has_correct_comments(scheduler['compute_l1_mod#compute_l1'].ir)
    assert has_correct_comments(scheduler['compute_l2_mod#compute_l2'].ir)
    assert has_correct_comments(scheduler['#another_l1'].ir)
    assert has_correct_comments(scheduler['#another_l2'].ir)

    # Rebuild the scheduler to wipe the previous result
    scheduler = Scheduler(
        paths=projA, includes=projA/'include', config=config,
        seed_routines='driverA', frontend=frontend, xmods=[tmp_path]
    )

    my_pipeline = partial(Pipeline, classes=(ZeroMyStuffTrafo, AddSnarkTrafo))
    # Then apply as a simple pipeline and check again
    scheduler.process(transformation=my_pipeline(name='Chad'))
    assert has_correct_assigns(scheduler['drivera_mod#drivera'].ir, 0)
    assert has_correct_assigns(scheduler['kernela_mod#kernela'].ir, 2)
    assert has_correct_assigns(scheduler['compute_l1_mod#compute_l1'].ir, 1)
    assert has_correct_assigns(scheduler['compute_l2_mod#compute_l2'].ir, 2, values=[66.0, 00])
    assert has_correct_assigns(scheduler['#another_l1'].ir, 1)
    assert has_correct_assigns(scheduler['#another_l2'].ir, 2, values=[77.0, 00])

    assert has_correct_comments(scheduler['drivera_mod#drivera'].ir, name='Chad')
    assert has_correct_comments(scheduler['kernela_mod#kernela'].ir, name='Chad')
    assert has_correct_comments(scheduler['compute_l1_mod#compute_l1'].ir, name='Chad')
    assert has_correct_comments(scheduler['compute_l2_mod#compute_l2'].ir, name='Chad')
    assert has_correct_comments(scheduler['#another_l1'].ir, name='Chad')
    assert has_correct_comments(scheduler['#another_l2'].ir, name='Chad')


def test_pipeline_config_compose(config):
    """
    Test the correct instantiation of a custom :any:`Pipeline`
    object from config.
    """
    my_config = config.copy()
    my_config['dimensions'] = {
        'horizontal': {'size': 'KLON', 'index': 'JL', 'bounds': ['KIDIA', 'KFDIA']},
        'block_dim': {'size': 'NGPBLKS', 'index': 'IBL'},
    }
    my_config['transformations'] = {
        'VectorWithTrim': {
            'classname': 'SCCVectorPipeline',
            'module': 'loki.transformations.single_column',
            'options': {
                'horizontal': '%dimensions.horizontal%',
                'block_dim': '%dimensions.block_dim%',
                'directive': 'openacc',
                'trim_vector_sections': True,
            },
        },
        'preprocess': {
            'classname': 'RemoveCodeTransformation',
            'module': 'loki.transformations',
            'options': {
                'call_names': 'dr_hook',
                'remove_imports': False
            }
        },
        'postprocess': {
            'classname': 'ModuleWrapTransformation',
            'module': 'loki.transformations.build_system',
            'options': {'module_suffix': '_module'}
        }
    }
    my_config['pipelines'] = {
        'MyVectorPipeline': {
            'transformations': [
                'preprocess',
                'VectorWithTrim',
                'postprocess',
            ],
        }
    }
    cfg = SchedulerConfig.from_dict(my_config)

    # Check that transformations and pipelines were created correctly
    assert cfg.transformations['VectorWithTrim']
    assert cfg.transformations['preprocess']
    assert cfg.transformations['postprocess']

    assert cfg.pipelines['MyVectorPipeline']
    pipeline = cfg.pipelines['MyVectorPipeline']
    assert isinstance(pipeline, Pipeline)

    # Check that the pipeline is correctly composed
    assert len(pipeline.transformations) == 9
    assert type(pipeline.transformations[0]).__name__ == 'RemoveCodeTransformation'
    assert type(pipeline.transformations[1]).__name__ == 'SCCFuseVerticalLoops'
    assert type(pipeline.transformations[2]).__name__ == 'SCCBaseTransformation'
    assert type(pipeline.transformations[3]).__name__ == 'SCCDevectorTransformation'
    assert type(pipeline.transformations[4]).__name__ == 'SCCDemoteTransformation'
    assert type(pipeline.transformations[5]).__name__ == 'SCCVecRevectorTransformation'
    assert type(pipeline.transformations[6]).__name__ == 'SCCAnnotateTransformation'
    assert type(pipeline.transformations[7]).__name__ == 'PragmaModelTransformation'
    assert type(pipeline.transformations[8]).__name__ == 'ModuleWrapTransformation'

    # Check for some specified and default constructor flags
    assert pipeline.transformations[0].call_names == ('dr_hook',)
    assert pipeline.transformations[0].remove_imports is False
    assert isinstance(pipeline.transformations[2].horizontal, Dimension)
    assert pipeline.transformations[2].horizontal.size == 'KLON'
    assert pipeline.transformations[2].horizontal.index == 'JL'
    assert pipeline.transformations[3].trim_vector_sections is True
    assert pipeline.transformations[8].replace_ignore_items is True

@pytest.fixture(name='multi_modes_config')
def fixture_multi_modes_config():
    return {
        'default': {'role': 'kernel', 'expand': True, 'strict': False, 'mode': 'm1', 'replicate': True},
        'routines': {
            'driver_0': {'role': 'driver', 'mode': 'm1', 'replicate': False},
            'driver_1': {'role': 'driver', 'mode': 'm1', 'replicate': False},
            'driver_2': {'role': 'driver', 'mode': 'm2', 'replicate': False},
            'driver_3': {'role': 'driver', 'mode': 'm3', 'replicate': False},
            'driver_4': {'role': 'driver', 'mode': 'm3', 'replicate': False},
            'nested_subroutine_3': {'ignore': ['test1', 'test2']}
        },
        'transformations': {
            'Idem1': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
            'Idem2': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
            'Idem3': {'classname': 'IdemTransformation', 'module': 'loki.transformations'}
        },
        'pipelines': {
            # NB: The pipeline name is important here since it determines the mode of the items
            #     that don't have an explicit mode set in the config (e.g. nested_subroutine_3)
            'm1': {'transformations': {'Idem1'}},
            'm2': {'transformations': {'Idem2'}},
            'm3': {'transformations': {'Idem3'}}
        },
    }


@pytest.fixture(name='check_multimodes_callgraph', scope='module')
def fixture_check_multimodes_callgraph():
    """
    Fixture that provides a function to check that the call graph of the scheduler
    is consistent with the provided callgraph dictionary
    """
    def check_callgraph(scheduler, callgraph):
        if not isinstance(callgraph, dict) or not callgraph:
            return
        for routine in callgraph.keys():
            routine_ir = scheduler[routine].ir
            imports = routine_ir.imports
            imported_symbols = ()
            for imp in imports:
                imported_symbols += imp.symbols
            successors = []
            for successor in callgraph[routine]:
                if '#' in successor:
                    successors.append(scheduler[successor].ir)
            successors_local_name = [str(successor.name).lower() for successor in successors]
            calls = FindNodes(ir.CallStatement).visit(routine_ir.body)
            for call in calls:
                assert str(call.name).lower() in successors_local_name
                assert str(call.name).lower() in imported_symbols
            check_callgraph(scheduler, callgraph[routine])
    yield check_callgraph


def test_scheduler_multi_modes_invalid_pipeline(testdir, tmp_path, multi_modes_config):
    """
    Test that the scheduler raises an error when a pipeline is missing for a mode
    """
    config = SchedulerConfig.from_dict(multi_modes_config)
    config.pipelines.pop('m1')

    proj_multi_mode = testdir/'sources/projMultiMode'
    builddir = tmp_path/'scheduler_multi_driver_modes_dir'
    builddir.mkdir(exist_ok=True)

    scheduler = Scheduler(paths=proj_multi_mode, config=config, xmods=[tmp_path], output_dir=builddir)
    with pytest.raises(RuntimeError):
        scheduler.process(config.pipelines, proc_strategy=ProcessingStrategy.PLAN)


@pytest.mark.parametrize('as_modules', [False, True])
@pytest.mark.parametrize('reinit_scheduler', [True, False])
def test_scheduler_multi_modes(testdir, tmp_path, multi_modes_config, reinit_scheduler,
                               as_modules, check_multimodes_callgraph):
    config = SchedulerConfig.from_dict(multi_modes_config)

    if as_modules:
        proj_multi_mode = testdir/'sources/projMultiModeModules'
    else:
        proj_multi_mode = testdir/'sources/projMultiMode'

    builddir = tmp_path/'scheduler_multi_driver_modes_dir'
    builddir.mkdir(exist_ok=True)

    scheduler = Scheduler(paths=proj_multi_mode, config=config, xmods=[tmp_path], output_dir=builddir)

    # We manually invoke the mode propagation to check that the correct modes are assigned
    scheduler._propagate_modes()
    expected_modes_per_item = {
        'driver_0_mod#driver_0': {'m1'},
        'driver_1_mod#driver_1': {'m1'},
        'driver_2_mod#driver_2': {'m2'},
        'driver_3_mod#driver_3': {'m3'},
        'driver_4_mod#driver_4': {'m3'},
        'subroutine_1_mod#subroutine_1': {'m1'},
        'subroutine_2_mod#subroutine_2': {'m2'},
        'subroutine_3_mod#subroutine_3': {'m1', 'm3'},
        'nested_subroutine_1_mod#nested_subroutine_1': {'m1', 'm3'},
        'nested_subroutine_2_mod#nested_subroutine_2': {'m2'},
        'nested_subroutine_3_mod#nested_subroutine_3': {'m1', 'm2', 'm3'}
    }

    if not as_modules:
        # strip scope prefixes for non-module items
        expected_modes_per_item = {
            (item[item.find('#'):] if 'driver' not in item else item): modes
            for item, modes in expected_modes_per_item.items()
        }

    assert {
        item.name: {item.mode} if 'driver' in item.name else item.trafo_data['inherited_mode']
        for item in scheduler.items
    } == expected_modes_per_item

    # Manually invoke the pipeline separation several times to ensure it is idempotent
    scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.PLAN)
    scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.PLAN)
    scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.PLAN)

    scheduler.process(config.pipelines, proc_strategy=ProcessingStrategy.PLAN)

    expected_items_per_mode = {
        'm1': {
            'driver_0_mod#driver_0',
            'driver_1_mod#driver_1',
            'subroutine_1_loki_m1_mod#subroutine_1_loki_m1',
            'subroutine_3_loki_m1_mod#subroutine_3_loki_m1',
            'nested_subroutine_1_loki_m1_mod#nested_subroutine_1_loki_m1',
            'nested_subroutine_3_loki_m1_mod#nested_subroutine_3_loki_m1',
        },
        'm2': {
            'driver_2_mod#driver_2',
            'subroutine_2_loki_m2_mod#subroutine_2_loki_m2',
            'nested_subroutine_2_loki_m2_mod#nested_subroutine_2_loki_m2',
            'nested_subroutine_3_loki_m2_mod#nested_subroutine_3_loki_m2',
        },
        'm3': {
            'driver_3_mod#driver_3',
            'driver_4_mod#driver_4',
            'subroutine_3_loki_m3_mod#subroutine_3_loki_m3',
            'nested_subroutine_1_loki_m3_mod#nested_subroutine_1_loki_m3',
            'nested_subroutine_3_loki_m3_mod#nested_subroutine_3_loki_m3',
        }
    }
    if not as_modules:
        expected_items_per_mode = {
            mod: {item[item.find('#'):] if 'driver' not in item else item for item in items}
            for mod, items in expected_items_per_mode.items()
        }

    items_per_mode = defaultdict(set)
    for item in scheduler.items:
        items_per_mode[item.mode].add(item.name)

    assert set(items_per_mode) == set(expected_items_per_mode)
    for _mode, _val in items_per_mode.items():
        assert expected_items_per_mode[_mode] == _val

    transformations = (
        ModuleWrapTransformation(module_suffix='_mod'),
        DependencyTransformation(suffix='_test', module_suffix='_mod'),
        FileWriteTransformation()
    )
    for transformation in transformations:
        scheduler.process(transformation, proc_strategy=ProcessingStrategy.PLAN)

    plan_trafo = CMakePlanTransformation(rootpath=proj_multi_mode)
    scheduler.process(transformation=plan_trafo, proc_strategy=ProcessingStrategy.PLAN)
    planfile = tmp_path/'planfile'
    plan_trafo.write_plan(planfile)

    # Validate the plan file content
    loki_plan = planfile.read_text()
    plan_pattern = re.compile(r'set\(\s*(\w+)\s*(.*?)\s*\)', re.DOTALL)
    plan_dict = {k: v.split() for k, v in plan_pattern.findall(loki_plan)}
    plan_dict = {k: {Path(s).stem for s in v} for k, v in plan_dict.items()}

    expected_keys = {'LOKI_SOURCES_TO_TRANSFORM', 'LOKI_SOURCES_TO_APPEND', 'LOKI_SOURCES_TO_REMOVE'}
    assert set(plan_dict.keys()) == expected_keys

    _expected_files_to_transform = {
        'driver_0_mod', 'driver_1_mod', 'driver_2_mod', 'driver_3_mod', 'driver_4_mod',
        'nested_subroutine_1_mod', 'nested_subroutine_2_mod', 'nested_subroutine_3_mod',
        'subroutine_1_mod', 'subroutine_2_mod', 'subroutine_3_mod'
    }
    if as_modules:
        expected_files_to_transform = _expected_files_to_transform
    else:
        expected_files_to_transform = {v.replace('_mod', '') if 'routine' in v else v
                                       for v in _expected_files_to_transform}
    _expected_files_to_append = {
        'driver_0_mod.m1', 'driver_1_mod.m1', 'driver_2_mod.m2', 'driver_3_mod.m3',
        'driver_4_mod.m3', 'nested_subroutine_1_loki_m1_mod.m1',
        'nested_subroutine_1_loki_m3_mod.m3', 'nested_subroutine_2_loki_m2_mod.m2',
        'nested_subroutine_3_loki_m1_mod.m1',
        'nested_subroutine_3_loki_m2_mod.m2', 'nested_subroutine_3_loki_m3_mod.m3',
        'subroutine_1_loki_m1_mod.m1', 'subroutine_2_loki_m2_mod.m2',
        'subroutine_3_loki_m1_mod.m1', 'subroutine_3_loki_m3_mod.m3'
    }
    if as_modules:
        expected_files_to_append = _expected_files_to_append
    else:
        expected_files_to_append = {v.replace('_mod', '') if 'routine' in v else v
                                    for v in _expected_files_to_append}
    expected_files_to_remove = {
        'driver_0_mod', 'driver_1_mod', 'driver_2_mod', 'driver_3_mod', 'driver_4_mod'
    }

    assert set(plan_dict['LOKI_SOURCES_TO_TRANSFORM']) == expected_files_to_transform
    assert set(plan_dict['LOKI_SOURCES_TO_APPEND']) == expected_files_to_append
    assert set(plan_dict['LOKI_SOURCES_TO_REMOVE']) == expected_files_to_remove

    if reinit_scheduler:
        scheduler = Scheduler(paths=proj_multi_mode, config=config, xmods=[tmp_path], output_dir=builddir)
    scheduler.propagate_and_separate_modes()
    for transformation in transformations:
        scheduler.process(transformation)

    expected_callgraph = {
        'driver_0_mod#driver_0': {
            'subroutine_1_loki_m1_test_mod#subroutine_1_loki_m1_test': {
                'nested_subroutine_1_loki_m1_test_mod#nested_subroutine_1_loki_m1_test',
            },
            'subroutine_3_loki_m1_test_mod#subroutine_3_loki_m1_test': {
                'nested_subroutine_1_loki_m1_test_mod#nested_subroutine_1_loki_m1_test',
                'nested_subroutine_3_loki_m1_test_mod#nested_subroutine_3_loki_m1_test',
            },
        },
        'driver_1_mod#driver_1': {
            'subroutine_1_loki_m1_test_mod#subroutine_1_loki_m1_test': {
                'nested_subroutine_1_loki_m1_test_mod#nested_subroutine_1_loki_m1_test',
            },
            'subroutine_3_loki_m1_test_mod#subroutine_3_loki_m1_test': {
                'nested_subroutine_1_loki_m1_test_mod#nested_subroutine_1_loki_m1_test',
                'nested_subroutine_3_loki_m1_test_mod#nested_subroutine_3_loki_m1_test',
            }
        },
        'driver_2_mod#driver_2': {
            'subroutine_2_loki_m2_test_mod#subroutine_2_loki_m2_test': {
                'nested_subroutine_2_loki_m2_test_mod#nested_subroutine_2_loki_m2_test',
                'nested_subroutine_3_loki_m2_test_mod#nested_subroutine_3_loki_m2_test'
            }
        },
        'driver_3_mod#driver_3': {
            'subroutine_3_loki_m3_test_mod#subroutine_3_loki_m3_test': {
                'nested_subroutine_1_loki_m3_test_mod#nested_subroutine_1_loki_m3_test',
                'nested_subroutine_3_loki_m3_test_mod#nested_subroutine_3_loki_m3_test',
            }
        },
        'driver_4_mod#driver_4': {
            'subroutine_3_loki_m3_test_mod#subroutine_3_loki_m3_test': {
                'nested_subroutine_1_loki_m3_test_mod#nested_subroutine_1_loki_m3_test',
                'nested_subroutine_3_loki_m3_test_mod#nested_subroutine_3_loki_m3_test',
            }
        }
    }

    check_multimodes_callgraph(scheduler, expected_callgraph)


@pytest.mark.parametrize('proc_strategy', [ProcessingStrategy.PLAN, ProcessingStrategy.DEFAULT])
def test_scheduler_multi_modes_imports(tmp_path, proc_strategy, check_multimodes_callgraph):
    """
    Test that the correct imports are added to the transformed files in multi-mode pipelines
    """

    fcode_driver1 = """
subroutine driver1
    use test_multi_modes_kernel1, only: kernel1, n
    use test_multi_modes_kernel2, only: kernel2
    implicit none
    integer :: arr(n, n)
    call kernel1
    call kernel2
end subroutine driver1
    """.strip()

    fcode_driver2 = """
subroutine driver2
    use test_multi_modes_kernel1, only: n
    use test_multi_modes_kernel2, only: kernel2
    implicit none
    integer :: arr(n, n)
    call kernel2
end subroutine driver2
    """.strip()

    fcode_kernel1 = """
module test_multi_modes_kernel1
implicit none
integer, parameter :: n = 10
contains
subroutine kernel1
    use test_multi_modes_kernel2, only: kernel2
    call kernel2
end subroutine kernel1
end module test_multi_modes_kernel1
    """.strip()

    fcode_kernel2 = """
module test_multi_modes_kernel2
implicit none
contains
subroutine kernel2
end subroutine kernel2
end module test_multi_modes_kernel2
    """.strip()

    config = {
        'default': {'role': 'kernel', 'expand': True, 'strict': False, 'mode': 'm1', 'replicate': True},
        'routines': {
            'driver1': {'role': 'driver', 'mode': 'm1', 'replicate': False},
            'driver2': {'role': 'driver', 'mode': 'm2', 'replicate': False},
        },
        'transformations': {
            'Idem1': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
            'Idem2': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
        },
        'pipelines': {
            'm1': {'transformations': {'Idem1'}},
            'm2': {'transformations': {'Idem2'}},
        },
    }
    scheduler_config = SchedulerConfig.from_dict(config)

    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    (src_dir/'driver1.F90').write_text(fcode_driver1)
    (src_dir/'driver2.F90').write_text(fcode_driver2)
    (src_dir/'test_multi_modes_kernel1.F90').write_text(fcode_kernel1)
    (src_dir/'test_multi_modes_kernel2.F90').write_text(fcode_kernel2)

    out_dir = tmp_path/'output'
    out_dir.mkdir()

    scheduler = Scheduler(paths=src_dir, config=scheduler_config, xmods=[tmp_path], output_dir=out_dir)
    scheduler.propagate_and_separate_modes(proc_strategy=proc_strategy)
    scheduler.process(scheduler_config.pipelines, proc_strategy=proc_strategy)

    if proc_strategy != ProcessingStrategy.PLAN:
        expected_dependencies = {
            '#driver1': {
                'test_multi_modes_kernel1', 'test_multi_modes_kernel1_loki_m1#kernel1_loki_m1',
                'test_multi_modes_kernel2_loki_m1#kernel2_loki_m1'
            },
            '#driver2': {'test_multi_modes_kernel1', 'test_multi_modes_kernel2_loki_m2#kernel2_loki_m2'},
            'test_multi_modes_kernel1_loki_m1#kernel1_loki_m1': {'test_multi_modes_kernel2_loki_m1#kernel2_loki_m1'},
        }

        check_multimodes_callgraph(scheduler, expected_dependencies)

        expected_imports = {
            '#driver1': {
                'test_multi_modes_kernel1', 'test_multi_modes_kernel1_loki_m1', 'test_multi_modes_kernel2_loki_m1'
            },
            '#driver2': {'test_multi_modes_kernel1', 'test_multi_modes_kernel2_loki_m2'},
            'test_multi_modes_kernel1_loki_m1#kernel1_loki_m1': {'test_multi_modes_kernel2_loki_m1'},
        }
        for item_name, imports in expected_imports.items():
            item_ir = scheduler[item_name].ir
            item_imports = item_ir.imports
            imported_modules = {imp.module for imp in item_imports}
            assert imported_modules == imports


def _write_shared_kernel_multi_mode_sources(src_dir):
    """
    Create a minimal multi-mode graph with two drivers sharing one kernel that
    calls a downstream leaf.
    """
    fcode_driver_small = """
subroutine driver_small
    use shared_kernel_mod, only: shared_kernel
    implicit none
    call shared_kernel
end subroutine driver_small
    """.strip()

    fcode_driver_stack = """
subroutine driver_stack
    use shared_kernel_mod, only: shared_kernel
    implicit none
    call shared_kernel
end subroutine driver_stack
    """.strip()

    fcode_shared_kernel = """
module shared_kernel_mod
implicit none
contains
subroutine shared_kernel
    use leaf_mod, only: leaf
    call leaf
end subroutine shared_kernel
end module shared_kernel_mod
    """.strip()

    fcode_leaf = """
module leaf_mod
implicit none
contains
subroutine leaf
end subroutine leaf
end module leaf_mod
    """.strip()

    (src_dir/'driver_small.F90').write_text(fcode_driver_small)
    (src_dir/'driver_stack.F90').write_text(fcode_driver_stack)
    (src_dir/'shared_kernel_mod.F90').write_text(fcode_shared_kernel)
    (src_dir/'leaf_mod.F90').write_text(fcode_leaf)


def _shared_kernel_multi_mode_config():
    return SchedulerConfig.from_dict({
        'default': {'role': 'kernel', 'expand': True, 'strict': False, 'mode': 'm_small', 'replicate': True},
        'routines': {
            'driver_small': {'role': 'driver', 'mode': 'm_small', 'replicate': False},
            'driver_stack': {'role': 'driver', 'mode': 'm_stack', 'replicate': False},
        },
        'transformations': {
            'IdemSmall': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
            'IdemStack': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
        },
        'pipelines': {
            'm_small': {'transformations': {'IdemSmall'}},
            'm_stack': {'transformations': {'IdemStack'}},
        },
    })


def _shared_kernel_multi_mode_override_config():
    return SchedulerConfig.from_dict({
        'default': {'role': 'kernel', 'expand': True, 'strict': False, 'mode': 'm_small', 'replicate': True},
        'routines': {
            'driver_small': {'role': 'driver', 'mode': 'm_small', 'replicate': False},
            'driver_stack': {'role': 'driver', 'mode': 'm_stack', 'replicate': False},
            'shared_kernel': {'role': 'kernel'},
            'leaf': {'role': 'kernel'},
        },
        'transformations': {
            'IdemSmall': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
            'IdemStack': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
        },
        'pipelines': {
            'm_small': {'transformations': {'IdemSmall'}},
            'm_stack': {'transformations': {'IdemStack'}},
        },
    })


def _shared_kernel_multi_mode_seed_override_config():
    return SchedulerConfig.from_dict({
        'default': {'role': 'kernel', 'expand': True, 'strict': False, 'mode': 'm_small', 'replicate': True},
        'routines': {
            'driver_small': {'role': 'driver', 'mode': 'm_small', 'replicate': False},
            'driver_stack': {'role': 'driver', 'mode': 'm_stack', 'replicate': False},
            'shared_kernel': {'role': 'kernel', 'seed_routine': True},
            'leaf': {'role': 'kernel', 'seed_routine': True},
        },
        'transformations': {
            'IdemSmall': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
            'IdemStack': {'classname': 'IdemTransformation', 'module': 'loki.transformations'},
        },
        'pipelines': {
            'm_small': {'transformations': {'IdemSmall'}},
            'm_stack': {'transformations': {'IdemStack'}},
        },
    })


def test_scheduler_multi_modes_shared_kernel_plan_variants(tmp_path):
    """
    Document current multi-mode planning behaviour for a shared kernel.

    Without explicit per-routine config entries, the original shared kernel is
    replaced entirely by renamed per-mode variants.
    """
    scheduler_config = _shared_kernel_multi_mode_config()

    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    _write_shared_kernel_multi_mode_sources(src_dir)

    out_dir = tmp_path/'output'
    out_dir.mkdir()

    scheduler = Scheduler(paths=src_dir, config=scheduler_config, xmods=[tmp_path], output_dir=out_dir)
    scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.PLAN)
    scheduler.process(scheduler_config.pipelines, proc_strategy=ProcessingStrategy.PLAN)
    scheduler.process(FileWriteTransformation(), proc_strategy=ProcessingStrategy.PLAN)

    plan_trafo = CMakePlanTransformation(rootpath=src_dir)
    scheduler.process(transformation=plan_trafo, proc_strategy=ProcessingStrategy.PLAN)
    planfile = tmp_path/'planfile'
    plan_trafo.write_plan(planfile)

    plan_pattern = re.compile(r'set\(\s*(\w+)\s*(.*?)\s*\)', re.DOTALL)
    plan_dict = {k: v.split() for k, v in plan_pattern.findall(planfile.read_text())}
    append_stems = [Path(s).stem for s in plan_dict['LOKI_SOURCES_TO_APPEND']]

    assert len(plan_dict['LOKI_SOURCES_TO_APPEND']) == len(set(plan_dict['LOKI_SOURCES_TO_APPEND']))

    assert 'shared_kernel_mod.m_small' not in append_stems
    assert append_stems.count('shared_kernel_loki_m_small_mod.m_small') == 1
    assert append_stems.count('shared_kernel_loki_m_stack_mod.m_stack') == 1

    assert 'leaf_mod.m_small' not in append_stems
    assert append_stems.count('leaf_loki_m_small_mod.m_small') == 1
    assert append_stems.count('leaf_loki_m_stack_mod.m_stack') == 1


def test_scheduler_multi_modes_shared_kernel_plan_variants_with_routine_override(tmp_path):
    """
    Explicit per-routine config entries change the current behaviour: the
    original shared kernel remains and same-mode clones are emitted in addition.
    """
    scheduler_config = _shared_kernel_multi_mode_override_config()

    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    _write_shared_kernel_multi_mode_sources(src_dir)

    out_dir = tmp_path/'output'
    out_dir.mkdir()

    scheduler = Scheduler(
        paths=src_dir, config=scheduler_config,
        seed_routines=['driver_small', 'driver_stack', 'shared_kernel', 'leaf'],
        xmods=[tmp_path], output_dir=out_dir
    )
    scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.PLAN)
    scheduler.process(scheduler_config.pipelines, proc_strategy=ProcessingStrategy.PLAN)
    scheduler.process(FileWriteTransformation(), proc_strategy=ProcessingStrategy.PLAN)

    plan_trafo = CMakePlanTransformation(rootpath=src_dir)
    scheduler.process(transformation=plan_trafo, proc_strategy=ProcessingStrategy.PLAN)
    planfile = tmp_path/'planfile'
    plan_trafo.write_plan(planfile)

    plan_pattern = re.compile(r'set\(\s*(\w+)\s*(.*?)\s*\)', re.DOTALL)
    plan_dict = {k: v.split() for k, v in plan_pattern.findall(planfile.read_text())}
    append_stems = [Path(s).stem for s in plan_dict['LOKI_SOURCES_TO_APPEND']]

    assert len(plan_dict['LOKI_SOURCES_TO_APPEND']) == len(set(plan_dict['LOKI_SOURCES_TO_APPEND']))

    # This matches the semantic duplication pattern seen in the real-world case.
    assert append_stems.count('shared_kernel_mod.m_small') == 1
    assert append_stems.count('shared_kernel_loki_m_small_mod.m_small') == 1
    assert append_stems.count('shared_kernel_loki_m_stack_mod.m_stack') == 1

    assert append_stems.count('leaf_mod.m_small') == 1
    assert append_stems.count('leaf_loki_m_small_mod.m_small') == 1
    assert append_stems.count('leaf_loki_m_stack_mod.m_stack') == 1


def test_scheduler_multi_modes_shared_kernel_call_import_rewrite_consistency(tmp_path):
    """
    Document how call and import rewrites currently behave for shared kernels.

    Both drivers are rewired to mode-specific clones, while the original shared
    kernel and leaf remain in the scheduler with their original call/import chain.
    """
    scheduler_config = _shared_kernel_multi_mode_override_config()

    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    _write_shared_kernel_multi_mode_sources(src_dir)

    out_dir = tmp_path/'output'
    out_dir.mkdir()

    scheduler = Scheduler(
        paths=src_dir, config=scheduler_config,
        seed_routines=['driver_small', 'driver_stack', 'shared_kernel', 'leaf'],
        xmods=[tmp_path], output_dir=out_dir
    )
    scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.DEFAULT)
    scheduler.process(scheduler_config.pipelines, proc_strategy=ProcessingStrategy.DEFAULT)

    expected_items = {
        '#driver_small',
        '#driver_stack',
        'shared_kernel_mod#shared_kernel',
        'shared_kernel_loki_m_small_mod#shared_kernel_loki_m_small',
        'shared_kernel_loki_m_stack_mod#shared_kernel_loki_m_stack',
        'leaf_mod#leaf',
        'leaf_loki_m_small_mod#leaf_loki_m_small',
        'leaf_loki_m_stack_mod#leaf_loki_m_stack',
    }
    assert expected_items <= {item.name for item in scheduler.items}

    def calls_for(item_name):
        routine = scheduler[item_name].ir
        return tuple(str(call.name).lower() for call in FindNodes(ir.CallStatement).visit(routine.body))

    def imports_for(item_name):
        routine = scheduler[item_name].ir
        return {imp.module.lower() for imp in routine.imports}

    assert calls_for('#driver_small') == ('shared_kernel_loki_m_small',)
    assert imports_for('#driver_small') == {'shared_kernel_loki_m_small_mod'}

    assert calls_for('#driver_stack') == ('shared_kernel_loki_m_stack',)
    assert imports_for('#driver_stack') == {'shared_kernel_loki_m_stack_mod'}

    assert calls_for('shared_kernel_loki_m_small_mod#shared_kernel_loki_m_small') == ('leaf_loki_m_small',)
    assert imports_for('shared_kernel_loki_m_small_mod#shared_kernel_loki_m_small') == {'leaf_loki_m_small_mod'}

    assert calls_for('shared_kernel_loki_m_stack_mod#shared_kernel_loki_m_stack') == ('leaf_loki_m_stack',)
    assert imports_for('shared_kernel_loki_m_stack_mod#shared_kernel_loki_m_stack') == {'leaf_loki_m_stack_mod'}

    # The original shared kernel remains active with the original import/call chain.
    assert calls_for('shared_kernel_mod#shared_kernel') == ('leaf',)
    assert imports_for('shared_kernel_mod#shared_kernel') == {'leaf_mod'}


def test_scheduler_multi_modes_explicit_routine_entries_do_not_become_seeds(tmp_path):
    """
    Explicit routine config entries alone do not become scheduler seeds when
    seed_routines is not provided.
    """
    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    _write_shared_kernel_multi_mode_sources(src_dir)

    no_override_scheduler = Scheduler(
        paths=src_dir, config=_shared_kernel_multi_mode_config(),
        xmods=[tmp_path], output_dir=tmp_path/'out-no-override'
    )
    override_scheduler = Scheduler(
        paths=src_dir, config=_shared_kernel_multi_mode_override_config(), xmods=[tmp_path],
        output_dir=tmp_path/'out-override'
    )

    assert no_override_scheduler.seeds == ('driver_small', 'driver_stack')
    assert override_scheduler.seeds == ('driver_small', 'driver_stack')

    no_override_scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.DEFAULT)
    override_scheduler.propagate_and_separate_modes(proc_strategy=ProcessingStrategy.DEFAULT)

    assert 'shared_kernel_mod#shared_kernel' not in {item.name for item in no_override_scheduler.items}
    assert 'shared_kernel_mod#shared_kernel' not in {item.name for item in override_scheduler.items}


def test_scheduler_implicit_seeds_include_seed_routine_entries(tmp_path):
    """
    Non-driver routines explicitly marked with ``seed_routine = true`` are added
    to the implicit seed set alongside driver routines.
    """
    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    _write_shared_kernel_multi_mode_sources(src_dir)

    scheduler = Scheduler(
        paths=src_dir, config=_shared_kernel_multi_mode_seed_override_config(),
        xmods=[tmp_path], output_dir=tmp_path/'output'
    )

    assert scheduler.seeds == ('driver_small', 'driver_stack', 'shared_kernel', 'leaf')


def test_scheduler_requires_implicit_seed_or_explicit_seed_routines(tmp_path, caplog):
    """
    When no explicit seeds are provided, the scheduler requires at least one
    implicit seed via a driver role or ``seed_routine = true``.
    """
    src_dir = tmp_path/'sources'
    src_dir.mkdir()
    _write_shared_kernel_multi_mode_sources(src_dir)

    config = SchedulerConfig.from_dict({
        'default': {'role': 'kernel', 'expand': True, 'strict': False},
        'routines': {
            'shared_kernel': {'role': 'kernel'},
            'leaf': {'role': 'kernel'},
        },
    })

    caplog.set_level(WARNING)

    # Once deprecated:
    # with pytest.raises(RuntimeError, match='No seed routines provided and no implicit seeds found'):
    scheduler = Scheduler(paths=src_dir, config=config, xmods=[tmp_path], output_dir=tmp_path/'output')
    # however, rn for backwards compability only expect warning and old behaviour
    assert set(scheduler.seeds) == {'shared_kernel', 'leaf'}
    assert len(caplog.records) == 1
    assert "Using implicit or unspecified seed routines is deprecated and will" in caplog.records[0].message



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Verify correct frontend behaviour for annotation-like IR nodes.
"""

import pytest

from loki import Module, Subroutine
from loki.expression import symbols as sym
from loki.frontend import available_frontends, OMNI, FP
from loki.ir import nodes as ir, FindNodes


@pytest.mark.parametrize('frontend', available_frontends())
def test_pragma_vs_comment(frontend, tmp_path):
    """
    Make sure pragmas and comments are identified correctly
    """
    fcode = """
module frontend_pragma_vs_comment
    implicit none
!$some pragma
    integer :: var1
!!$some comment
    integer :: var2
!some comment
    integer :: var3
    !$some pragma
    integer :: var4
    ! !$some comment
    integer :: var5
end module frontend_pragma_vs_comment
    """.strip()

    module = Module.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    pragmas = FindNodes(ir.Pragma).visit(module.ir)
    comments = FindNodes(ir.Comment).visit(module.ir)
    assert len(pragmas) == 2
    assert len(comments) == 3
    assert all(pragma.keyword == 'some' for pragma in pragmas)
    assert all(pragma.content == 'pragma' for pragma in pragmas)
    assert all('some comment' in comment.text for comment in comments)


@pytest.mark.parametrize('frontend', available_frontends())
def test_pragma_line_continuation(frontend):
    """
    Test that multi-line pragmas are parsed and dealt with correctly.
    """
    fcode = """
SUBROUTINE TOTO(A,B)

IMPLICIT NONE
REAL, INTENT(IN) :: A
REAL, INTENT(INOUT) :: B

!$ACC PARALLEL LOOP GANG &
!$ACC& PRESENT(ZRDG_LCVQ,ZFLU_QSATS,ZRDG_CVGQ) &
!$ACC& PRIVATE (JBLK) &
!$ACC& VECTOR_LENGTH (YDCPG_OPTS%KLON)
!$ACC SEQUENTIAL

END SUBROUTINE TOTO
"""
    routine = Subroutine.from_source(fcode, frontend=frontend)

    pragmas = FindNodes(ir.Pragma).visit(routine.body)

    assert len(pragmas) == 2
    assert pragmas[0].keyword == 'ACC'
    assert 'PARALLEL' in pragmas[0].content
    assert 'PRESENT' in pragmas[0].content
    assert 'PRIVATE' in pragmas[0].content
    assert 'VECTOR_LENGTH' in pragmas[0].content
    assert pragmas[1].content == 'SEQUENTIAL'

    # Check that source object was generated right
    assert pragmas[0].source
    assert pragmas[0].source.lines == (8, 8) if frontend == OMNI else (8, 11)
    assert pragmas[1].source
    assert pragmas[1].source.lines == (12, 12)


@pytest.mark.parametrize('frontend', available_frontends())
def test_comment_block_clustering(frontend):
    """
    Test that multiple :any:`Comment` nodes into a :any:`CommentBlock`.
    """
    fcode = """
subroutine test_comment_block(a, b)
  ! What is this?
  ! Ohhh, ... a docstring?
  real, intent(inout) :: a, b

  a = a + 1.0
  ! Never gonna
  b = b + 2
  ! give you
  ! up...

  a = a + b
  ! Shut up, ...
  ! Rick!
end subroutine test_comment_block
"""
    routine = Subroutine.from_source(fcode, frontend=frontend)

    comments = FindNodes(ir.Comment).visit(routine.spec)
    assert len(comments) == 0
    blocks = FindNodes(ir.CommentBlock).visit(routine.spec)
    assert len(blocks) == 0

    assert isinstance(routine.docstring[0], ir.CommentBlock)
    assert len(routine.docstring[0].comments) == 2
    assert routine.docstring[0].comments[0].text == '! What is this?'
    assert routine.docstring[0].comments[1].text == '! Ohhh, ... a docstring?'

    comments = FindNodes(ir.Comment).visit(routine.body)
    assert len(comments) == 2 if frontend == FP else 1
    assert comments[-1].text == '! Never gonna'

    blocks = FindNodes(ir.CommentBlock).visit(routine.body)
    assert len(blocks) == 2
    assert len(blocks[0].comments) == 3 if frontend == FP else 2
    assert blocks[0].comments[0].text == '! give you'
    assert blocks[0].comments[1].text == '! up...'

    assert len(blocks[1].comments) == 2
    assert blocks[1].comments[0].text == '! Shut up, ...'
    assert blocks[1].comments[1].text == '! Rick!'


@pytest.mark.parametrize('frontend', available_frontends(
    skip=[(OMNI, 'OMNI strips comments during parse')]
))
def test_inline_comments(frontend):
    """
    Test that multiple :any:`Comment` nodes into a :any:`CommentBlock`.
    """
    fcode = """
subroutine test_inline_comments(a, b)
  real, intent(inout) :: a, b  ! We don't need no education
  real, external :: alien_func ! We don't need no thought control
  integer :: i

  a = a + 1.0
  ! Who said that?
  b = b + 2             ! All in all it's just another

  do i=1, 10
    b = b + 2           ! Brick in the ...
  enddo

  a = a + alien_func()  ! wall !
end subroutine test_inline_comments
"""
    routine = Subroutine.from_source(fcode, frontend=frontend)

    decls = FindNodes(ir.VariableDeclaration).visit(routine.spec)
    assert len(decls) == 2
    assert decls[0].comment.text == "! We don't need no education"
    assert decls[1].comment is None

    proc_decls = FindNodes(ir.ProcedureDeclaration).visit(routine.spec)
    assert len(proc_decls) == 1
    assert proc_decls[0].comment.text == "! We don't need no thought control"

    assigns = FindNodes(ir.Assignment).visit(routine.body)
    assert len(assigns) == 4
    assert assigns[0].comment is None
    assert assigns[1].comment.text == "! All in all it's just another"
    assert assigns[2].comment.text == '! Brick in the ...'
    assert assigns[3].comment.text == '! wall !'

    comments = FindNodes(ir.Comment).visit(routine.body)
    assert len(comments) == 4
    assert comments[1].text == '! Who said that?'
    assert comments[0].text == comments[2].text == comments[3].text == ''


@pytest.mark.parametrize('frontend', available_frontends(skip=[(OMNI, 'OMNI does not like Loki pragmas, yet!')]))
def test_routine_variables_dimension_pragmas(frontend):
    """
    Test that `!$loki dimension` pragmas can be used to override the
    conceptual `.shape` of local and argument variables.
    """
    fcode = """
subroutine routine_variables_dimensions(x, y, v0, v1, v2, v3, v4)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y

  !$loki dimension(10)
  real(kind=jprb), intent(inout) :: v0(:)
  !$loki dimension(x)
  real(kind=jprb), intent(inout) :: v1(:)
  !$loki dimension(x,y,:)
  real(kind=jprb), dimension(:,:,:), intent(inout) :: v2, v3
  !$loki dimension(x,y)
  real(kind=jprb), pointer, intent(inout) :: v4(:,:)
  !$loki dimension(x+y,2*x)
  real(kind=jprb), allocatable :: v5(:,:)
  !$loki dimension(x/2, x**2, (x+y)/x)
  real(kind=jprb), dimension(:, :, :), pointer :: v6

end subroutine routine_variables_dimensions
"""

    def to_str(expr):
        return str(expr).lower().replace(' ', '')

    routine = Subroutine.from_source(fcode, frontend=frontend)
    assert routine.variable_map['v0'].shape[0] == 10
    assert isinstance(routine.variable_map['v0'].shape[0], sym.IntLiteral)
    assert isinstance(routine.variable_map['v1'].shape[0], sym.Scalar)
    assert routine.variable_map['v2'].shape[0] == 'x'
    assert routine.variable_map['v2'].shape[1] == 'y'
    assert routine.variable_map['v2'].shape[2] == ':'
    assert isinstance(routine.variable_map['v2'].shape[0], sym.Scalar)
    assert isinstance(routine.variable_map['v2'].shape[1], sym.Scalar)
    assert isinstance(routine.variable_map['v2'].shape[2], sym.RangeIndex)
    assert routine.variable_map['v3'].shape[0] == 'x'
    assert routine.variable_map['v3'].shape[1] == 'y'
    assert routine.variable_map['v3'].shape[2] == ':'
    assert isinstance(routine.variable_map['v3'].shape[0], sym.Scalar)
    assert isinstance(routine.variable_map['v3'].shape[1], sym.Scalar)
    assert isinstance(routine.variable_map['v3'].shape[2], sym.RangeIndex)
    assert routine.variable_map['v4'].shape[0] == 'x'
    assert routine.variable_map['v4'].shape[1] == 'y'
    assert isinstance(routine.variable_map['v4'].shape[0], sym.Scalar)
    assert isinstance(routine.variable_map['v4'].shape[1], sym.Scalar)
    assert to_str(routine.variable_map['v5'].shape[0]) == 'x+y'
    assert to_str(routine.variable_map['v5'].shape[1]) == '2*x'
    assert isinstance(routine.variable_map['v5'].shape[0], sym.Sum)
    assert isinstance(routine.variable_map['v5'].shape[1], sym.Product)
    assert to_str(routine.variable_map['v6'].shape[0]) == 'x/2'
    assert to_str(routine.variable_map['v6'].shape[1]) == 'x**2'
    assert to_str(routine.variable_map['v6'].shape[2]) == '(x+y)/x'
    assert isinstance(routine.variable_map['v6'].shape[0], sym.Quotient)
    assert isinstance(routine.variable_map['v6'].shape[1], sym.Power)
    assert isinstance(routine.variable_map['v6'].shape[2], sym.Quotient)


@pytest.mark.parametrize('frontend', available_frontends(skip=[(OMNI, 'OMNI does not like Loki pragmas, yet!')]))
def test_module_variables_dimension_pragmas(frontend, tmp_path):
    """
    Test that `!$loki dimension` pragmas can be used to override the
    conceptual `.shape` of module variables.
    """
    code_mod = """
module mod_variable_dimensions

  integer, parameter :: jprb = selected_real_kind(13,300)
  integer :: x, y

  !$loki dimension(10)
  real(kind=jprb), intent(inout) :: v0(:)
  !$loki dimension(x)
  real(kind=jprb), intent(inout) :: v1(:)
  !$loki dimension(x,y,:)
  real(kind=jprb), dimension(:,:,:), intent(inout) :: v2, v3
  !$loki dimension(x,y)
  real(kind=jprb), pointer, intent(inout) :: v4(:,:)
  !$loki dimension(x+y,2*x)
  real(kind=jprb), allocatable :: v5(:,:)
  !$loki dimension(x/2, x**2, (x+y)/x)
  real(kind=jprb), dimension(:, :, :), pointer :: v6
end module mod_variable_dimensions
    """

    def to_str(expr):
        return str(expr).lower().replace(' ', '')

    mod = Module.from_source(code_mod, frontend=frontend, xmods=[tmp_path])
    variable_map = mod.variable_map
    assert variable_map['v0'].shape[0] == 10
    assert isinstance(variable_map['v0'].shape[0], sym.IntLiteral)
    assert isinstance(variable_map['v1'].shape[0], sym.Scalar)
    assert variable_map['v2'].shape[0] == 'x'
    assert variable_map['v2'].shape[1] == 'y'
    assert variable_map['v2'].shape[2] == ':'
    assert isinstance(variable_map['v2'].shape[0], sym.Scalar)
    assert isinstance(variable_map['v2'].shape[1], sym.Scalar)
    assert isinstance(variable_map['v2'].shape[2], sym.RangeIndex)
    assert variable_map['v3'].shape[0] == 'x'
    assert variable_map['v3'].shape[1] == 'y'
    assert variable_map['v3'].shape[2] == ':'
    assert isinstance(variable_map['v3'].shape[0], sym.Scalar)
    assert isinstance(variable_map['v3'].shape[1], sym.Scalar)
    assert isinstance(variable_map['v3'].shape[2], sym.RangeIndex)
    assert variable_map['v4'].shape[0] == 'x'
    assert variable_map['v4'].shape[1] == 'y'
    assert isinstance(variable_map['v4'].shape[0], sym.Scalar)
    assert isinstance(variable_map['v4'].shape[1], sym.Scalar)
    assert to_str(variable_map['v5'].shape[0]) == 'x+y'
    assert to_str(variable_map['v5'].shape[1]) == '2*x'
    assert isinstance(variable_map['v5'].shape[0], sym.Sum)
    assert isinstance(variable_map['v5'].shape[1], sym.Product)
    assert to_str(variable_map['v6'].shape[0]) == 'x/2'
    assert to_str(variable_map['v6'].shape[1]) == 'x**2'
    assert to_str(variable_map['v6'].shape[2]) == '(x+y)/x'
    assert isinstance(variable_map['v6'].shape[0], sym.Quotient)
    assert isinstance(variable_map['v6'].shape[1], sym.Power)
    assert isinstance(variable_map['v6'].shape[2], sym.Quotient)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Collection of classes to represent type information for symbols used throughout
:doc:`internal_representation`
"""

from enum import Enum

from loki.tools import flatten


__all__ = [
    'DataType', 'BasicType',
    'DEFERRED', 'LOGICAL', 'INTEGER', 'REAL', 'CHARACTER', 'COMPLEX'
]


class DataType:
    """
    Base class for data types a symbol may have
    """


class BasicType(DataType, int, Enum):
    """
    Representation of intrinsic data types, names taken from the FORTRAN convention.

    Currently, there are

    - :any:`LOGICAL`
    - :any:`INTEGER`
    - :any:`REAL`
    - :any:`CHARACTER`
    - :any:`COMPLEX`

    and, to indicate an undefined data type (e.g., for imported
    symbols whose definition is not available), :any:`DEFERRED`.

    For convenience, string representations of FORTRAN and C99 types can be
    heuristically converted.
    """

    DEFERRED = 0
    LOGICAL = 1
    INTEGER = 2
    REAL = 3
    CHARACTER = 4
    COMPLEX = 5

    @classmethod
    def from_str(cls, value):
        """
        Try to convert the given string using one of the `from_*` methods.
        """
        lookup_methods = (cls.from_name, cls.from_fortran_type, cls.from_c99_type)
        for meth in lookup_methods:
            try:
                return meth(value)
            except KeyError:
                pass
        raise ValueError(f'Unknown data type: {value}')

    @classmethod
    def from_name(cls, value):
        """
        Convert the given string representation of the :any:`BasicType`.
        """
        return {t.name: t for t in cls}[value]

    @classmethod
    def from_fortran_type(cls, value):
        """
        Convert the given string representation of a FORTRAN type.
        """
        type_map = {'logical': cls.LOGICAL, 'integer': cls.INTEGER, 'real': cls.REAL,
                    'double precision': cls.REAL, 'double complex': cls.COMPLEX,
                    'character': cls.CHARACTER, 'complex': cls.COMPLEX}
        return type_map[value.lower()]

    @classmethod
    def from_c99_type(cls, value):
        """
        Convert the given string representation of a C99 type.
        """
        logical_types = ['bool', '_Bool']
        integer_types = ['short', 'int', 'long', 'long long']
        integer_types += flatten([(f'signed {t}', f'unsigned {t}') for t in integer_types])
        real_types = ['float', 'double', 'long double']
        character_types = ['char']
        complex_types = ['float _Complex', 'double _Complex', 'long double _Complex']

        type_map = {t: cls.LOGICAL for t in logical_types}
        type_map.update({t: cls.INTEGER for t in integer_types})
        type_map.update({t: cls.REAL for t in real_types})
        type_map.update({t: cls.CHARACTER for t in character_types})
        type_map.update({t: cls.COMPLEX for t in complex_types})

        return type_map[value]


DEFERRED = BasicType.DEFERRED
LOGICAL = BasicType.LOGICAL
INTEGER = BasicType.INTEGER
REAL = BasicType.REAL
CHARACTER = BasicType.CHARACTER
COMPLEX = BasicType.COMPLEX



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

""" Generic and specific statement node type definitions. """

from dataclasses import field
from typing import Optional, Union, Tuple

from pymbolic.primitives import Expression
from pydantic import field_validator, model_validator, ValidationError
from pydantic_core import ArgsKwargs

from loki.ir.nodes.abstract_nodes import LeafNode
from loki.tools import dataclass_strict, truncate_string, sanitize_tuple


__all__ = [
    'GenericStmt', 'ImplicitStmt', 'SaveStmt', 'PublicStmt',
    'PrivateStmt', 'CommonStmt', 'ContainsStmt', 'ReturnStmt',
    'CycleStmt', 'ContinueStmt', 'StopStmt', 'ExitStmt', 'GotoStmt',
    'PrintStmt', 'FormatStmt'
]


@dataclass_strict(frozen=True)
class _GenericStmtBase():
    """ Type definitions for :any:`GenericStmt` node type. """

    text: str


@dataclass_strict(frozen=True)
class GenericStmt(LeafNode, _GenericStmtBase):
    """
    Catch-all generic node for corner-cases.

    This is provided as a fallback for any statements that do not have
    an appropriate representation in the IR. These can either be language
    features for which support was not yet added, or statements that are not
    relevant in Loki's scope of applications. This node retains the text of
    the statement in the original source as-is.

    Parameters
    ----------
    text : str
        The statement as a string.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = None

    def __repr__(self):
        return f'GenericStmt:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class ImplicitStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``IMPLICIT`` statement.

    Parameters
    ----------
    text : str or :any:`Expression`, optional
        Either a tuple of variable specifiers or a string; default: ``NONE``
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'IMPLICIT'

    text: Optional[Union[str, Tuple[Expression, ...]]] = 'NONE'

    @field_validator('text', mode='before')
    @classmethod
    def ensure_str_or_tuple(cls, value):
        if isinstance(value, str):
            return value
        return sanitize_tuple(value)

    def __repr__(self):
        return f'Implicit:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class SaveStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``SAVE`` statement.

    Parameters
    ----------
    text : str or :any:`Expression`, optional
        Either a tuple of variable specifiers or a string; default: ``NONE``
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'SAVE'

    text: Optional[Tuple[Expression, ...]] = ()

    @field_validator('text', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    def __repr__(self):
        return f'Save:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class PublicStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``PUBLIC`` specifier.

    Parameters
    ----------
    text : str or tuple of :any:`Expression`, optional
        Either a tuple of variable specifiers or a string; default: ``NONE``
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'PUBLIC'

    text: Optional[Tuple[Expression, ...]] = ()

    @field_validator('text', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    def __repr__(self):
        return f'Public:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class PrivateStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``PRIVATE`` specifier.

    Parameters
    ----------
    text : str or tuple of :any:`Expression`, optional
        Either a tuple of variable specifiers or a string; default: ``NONE``
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'PRIVATE'

    text: Optional[Tuple[Expression, ...]] = ()

    @field_validator('text', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    def __repr__(self):
        return f'Private:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class CommonStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``COMMON`` specifier.

    Parameters
    ----------
    text : str or tuple of :any:`Expression`, optional
        Either a tuple of variable specifiers or a string; default: ``NONE``
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'COMMON'

    text: Tuple[Expression, ...]

    @field_validator('text', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    def __repr__(self):
        return f'Common:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class ContainsStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``CONTAINS`` specifier.

    Parameters
    ----------
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'CONTAINS'

    text: Optional[None] = None

    def __post_init__(self):
        super().__post_init__()
        if not self.text is None:
            raise ValidationError('[Loki] ContainsStmt takes no constructor arguments')

    def __repr__(self):
        return 'Contains::'


@dataclass_strict(frozen=True)
class ReturnStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``RETURN`` specifier.

    Parameters
    ----------
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'RETURN'

    text: Optional[None] = None

    def __post_init__(self):
        super().__post_init__()
        if not self.text is None:
            raise ValidationError('[Loki] ReturnStmt takes no constructor arguments')

    def __repr__(self):
        return 'Return::'


@dataclass_strict(frozen=True)
class CycleStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``CYCLE`` specifier.

    Parameters
    ----------
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'CYCLE'

    text: Optional[None] = None

    def __post_init__(self):
        super().__post_init__()
        if not self.text is None:
            raise ValidationError('[Loki] CycleStmt takes no constructor arguments')

    def __repr__(self):
        return 'Cycle::'


@dataclass_strict(frozen=True)
class ContinueStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``CONTINUE`` specifier.

    Parameters
    ----------
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'CONTINUE'

    text: str = ''

    def __post_init__(self):
        super().__post_init__()
        if self.text:
            raise ValidationError('[Loki] ContinueStmt takes no constructor arguments')

    def __repr__(self):
        return 'Continue::'


@dataclass_strict(frozen=True)
class StopStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``STOP`` specifier.

    Parameters
    ----------
    text : str, optional
        The optional error code (integer or character) returned on stopping.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'STOP'

    text: Optional[Union[Expression, str]] = ''

    @field_validator('text', mode='before')
    @classmethod
    def ensure_literal(cls, value):
        from loki.expression import symbols as sym  # pylint: disable=import-outside-toplevel
        return sym.Literal(value) if value else ''

    def __repr__(self):
        return f'Stop::{self.text if self.text else ""}'


@dataclass_strict(frozen=True)
class ExitStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``EXIT`` specifier.

    Parameters
    ----------
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'EXIT'

    text: Optional[Union[Expression, str]] = ''

    @field_validator('text', mode='before')
    @classmethod
    def ensure_literal(cls, value):
        from loki.expression import symbols as sym  # pylint: disable=import-outside-toplevel
        return sym.IntLiteral(value) if value else ''

    def __repr__(self):
        return f'Exit::{self.text if self.text else ""}'


@dataclass_strict(frozen=True)
class GotoStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``GO TO`` specifier.

    Parameters
    ----------
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'GO TO'

    text: str

    def __repr__(self):
        return f'Goto:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class PrintStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``PRINT`` specifier.

    Parameters
    ----------
    values : tuple of str or :any:`Expression`
        Tuple of print specifier and variable specifiers.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'PRINT'

    # For output statements, the ``text`` is derived from a tuple of ``values``
    text: str = field(init=False)  # pylint: disable=invalid-field-call
    values: Tuple[Union[Expression, str], ...] = ()

    @model_validator(mode='before')
    @classmethod
    def disallow_text_argument(cls, data):
        """ Ensure the alternating construcotr arguments are used. """
        kwargs = getattr(data, 'kwargs', None) or {}
        if 'text' in kwargs:
            raise ValueError('PrintStmt uses values= for constructor input; text is derived from values')
        if not getattr(data, 'args', None) and 'values' not in kwargs:
            raise ValueError('PrintStmt requires at least one value')
        if getattr(data, 'args', None):
            return ArgsKwargs((), {'values': data.args[0], **kwargs})
        return data

    @field_validator('values', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    @field_validator('values')
    @classmethod
    def require_values(cls, value):
        if not value:
            raise ValueError('PrintStmt requires at least one value')
        return value

    def __post_init__(self):
        super().__post_init__()
        object.__setattr__(self, 'text', ', '.join(str(v) for v in self.values))

    @property
    def args(self):
        """ Ensure safe cloning by excluding derived ``text`` property from ``args``. """
        return {k: v for k, v in super().args.items() if k != 'text'}

    def __repr__(self):
        return f'Print:: {truncate_string(self.text)}'


@dataclass_strict(frozen=True)
class FormatStmt(GenericStmt):
    """
    :any:`GenericStmt` node that represents the ``FORMAT`` specifier.

    Parameters
    ----------
    values : tuple of str or :any:`Expression`, optional
        Tuple of format specifiers.
    **kwargs : optional
        Other parameters that are passed on to the parent class constructor.
    """

    keyword = 'FORMAT'

    # For output statements, the ``text`` is derived from a tuple of ``values``
    text: str = field(init=False)  # pylint: disable=invalid-field-call
    values: Optional[Tuple[Union[Expression, str], ...]] = ()

    @model_validator(mode='before')
    @classmethod
    def disallow_text_argument(cls, data):
        """ Ensure the alternating construcotr arguments are used. """
        kwargs = getattr(data, 'kwargs', None) or {}
        if 'text' in kwargs:
            raise ValueError('FormatStmt uses values= for constructor input; text is derived from values')
        if getattr(data, 'args', None):
            return ArgsKwargs((), {'values': data.args[0], **kwargs})
        return data

    @field_validator('values', mode='before')
    @classmethod
    def ensure_tuple(cls, value):
        return sanitize_tuple(value)

    def __post_init__(self):
        super().__post_init__()
        object.__setattr__(self, 'text', ', '.join(str(v) for v in self.values))

    @property
    def args(self):
        """ Ensure safe cloning by excluding derived ``text`` property from ``args``. """
        return {k: v for k, v in super().args.items() if k != 'text'}

    def __repr__(self):
        return f'Format:: {truncate_string(self.text)}'



# (C) Copyright 2024- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any

from loki.ir import Transformer

__all__ = ['AbstractDataflowAnalysis', 'dfa_attached']


class AbstractDataflowAnalysis(ABC):
    """
    Base interface for dataflow analyses attaching metadata to IR nodes.

    This abstract class defines the common entry points for providing
    "Attacher" and "Detacher" :any:`Transformation` objects that
    populate the implicit sets of symbols attached to each IR node
    when the analysis has been performed.

    For more info, see the :any:`dfa_attached` context.
    """

    class _Attacher(Transformer):
        """Default base class for analysis attachers."""

    class _Detacher(Transformer):
        """Default base class for analysis detachers."""

    def get_attacher(self) -> Any:
        """ Returns an instance of the associated "Attacher" :any:`Transformation`. """
        return self._Attacher()

    def get_detacher(self) -> Any:
        """ Returns an instance of the associated "Detacher" :any:`Transformation`. """
        return self._Detacher()

    @abstractmethod
    def attach_dataflow_analysis(self, module_or_routine):
        pass

    @abstractmethod
    def detach_dataflow_analysis(self, module_or_routine):
        pass


@contextmanager
def dfa_attached(module_or_routine, dfa):
    """
    Create a context in which information about defined, live and used symbols
    is attached to each IR node.

    This makes for each IR node the following properties available:

    * :attr:`Node.live_symbols`: symbols defined before the node;
    * :attr:`Node.defines_symbols`: symbols (potentially) defined by the
      node;
    * :attr:`Node.uses_symbols`: symbols used by the node that had to be
      defined before.

    This is an in-place update of nodes and thus existing references to IR
    nodes remain valid. When leaving the context the information is removed
    from IR nodes, while existing references remain valid.

    The default analysis, which uses a simplified Reaching Definitions
    Analysis (see :any:`DataflowAnalysis`) is used, unless a more
    specialised implementation is provided via the ``dfa`` argument.

    Parameters
    ----------
    module_or_routine : :any:`Module` or :any:`Subroutine`
        The object for which the IR is to be annotated.
    dfa : :any:`AbstractDataflowAnalysis`, optional
        Instance of a dataflow analysis object providing "attacher"
        and "detacher" mechanisms.
    """
    dfa.attach_dataflow_analysis(module_or_routine)
    try:
        yield module_or_routine
    finally:
        dfa.detach_dataflow_analysis(module_or_routine)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
from pymbolic.primitives import Expression

from loki import Module, Subroutine
from loki.expression import (
    symbols as sym, ExpressionCallbackMapper, ExpressionRetriever
)
from loki.frontend import available_frontends, OMNI
from loki.ir import (
    nodes as ir, is_parent_of, is_child_of, FindNodes, FindScopes,
    FindVariables, ExpressionFinder
)
from loki.tools import OrderedSet


@pytest.mark.parametrize('frontend', available_frontends())
def test_find_nodes_greedy(frontend):
    """
    Test the FindNodes visitor's greedy property.
    """
    fcode = """
subroutine routine_find_nodes_greedy(n, m)
  integer, intent(in) :: n, m

  if (n > m) then
    if (n == 3) then
      print *,"Inner if"
    endif
    print *,"Outer if"
  endif
end subroutine routine_find_nodes_greedy
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    conditionals = FindNodes(ir.Conditional).visit(routine.body)
    assert len(conditionals) == 2

    outer_cond = FindNodes(ir.Conditional, greedy=True).visit(routine.body)
    assert len(outer_cond) == 1
    assert outer_cond[0] in conditionals
    assert str(outer_cond[0].condition) == 'n > m'


@pytest.mark.parametrize('frontend', available_frontends())
def test_find_scopes(frontend):
    """
    Test the FindScopes visitor.
    """
    fcode = """
subroutine routine_find_nodes_greedy(n, m)
  integer, intent(in) :: n, m

  if (n > m) then
    if (n == 3) then
      print *,"Inner if"
    endif
    print *,"Outer if"
  endif
end subroutine routine_find_nodes_greedy
""".strip()
    routine = Subroutine.from_source(fcode, frontend=frontend)

    stmts = FindNodes(ir.GenericStmt).visit(routine.body)
    assert len(stmts) == 2
    inner = [i for i in stmts if 'Inner' in i.text][0]
    outer = [i for i in stmts if 'Outer' in i.text][0]

    conditionals = FindNodes(ir.Conditional).visit(routine.body)
    assert len(conditionals) == 2

    scopes = FindScopes(inner).visit(routine.body)
    assert len(scopes) == 1  # returns a list containing a list of nested nodes
    assert len(scopes[0]) == 4  # should have found 3 scopes and the node itself
    assert all(c in scopes[0] for c in conditionals)  # should have found all if
    assert routine.body is scopes[0][0]  # body section should be outermost scope
    assert str(scopes[0][1].condition) == 'n > m'  # outer if should come first
    assert inner is scopes[0][-1]  # node itself should be last in list

    scopes = FindScopes(outer).visit(routine.body)
    assert len(scopes) == 1  # returns a list containing a list of nested nodes
    assert len(scopes[0]) == 3  # should have found 2 scopes and the node itself
    assert all(c in scopes[0] or str(c.condition == 'n == 3')
               for c in conditionals)  # should have found only the outer if
    assert routine.body is scopes[0][0]  # body section should be outermost scope
    assert outer is scopes[0][-1]  # node itself should be last in list


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_finder(frontend):
    """
    Test the expression finder's ability to yield only all variables.
    """
    fcode = """
subroutine routine_simple (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i

  do i=1, x
     vector(i) = vector(i) + scalar
     matrix(i, :) = i * vector(i)
  end do
end subroutine routine_simple
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    variables = FindVariables(unique=False).visit(routine.body)
    assert len(variables) == 12
    assert all(isinstance(v, Expression) for v in variables)

    assert sorted([str(v) for v in variables]) == (
        ['i'] * 6 + ['matrix(i, :)', 'scalar'] + ['vector(i)'] * 3 + ['x'])


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_finder_unique(frontend):
    """
    Test the expression finder's ability to yield unique variables.
    """
    fcode = """
subroutine routine_simple (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i

  do i=1, x
     vector(i) = vector(i) + scalar
     matrix(i, :) = i * vector(i)
  end do
end subroutine routine_simple
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    variables = FindVariables().visit(routine.body)
    assert isinstance(variables, OrderedSet)
    assert len(variables) == 5
    assert all(isinstance(v, Expression) for v in variables)

    assert sorted([str(v) for v in variables]) == ['i', 'matrix(i, :)', 'scalar', 'vector(i)', 'x']


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_finder_with_ir_node(frontend):
    """
    Test the expression finder's ability to yield the root node.
    """
    fcode = """
subroutine routine_simple (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i

  do i=1, x
     vector(i) = vector(i) + scalar
     matrix(i, :) = i * vector(i)
  end do
end subroutine routine_simple
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    variables = FindVariables(unique=False, with_ir_node=True).visit(routine.body)
    assert len(variables) == 3
    assert all(isinstance(v, tuple) and len(v) == 2 for v in variables)

    # Verify that the variables in the loop definition are found
    loops = [v for v in variables if isinstance(v[0], ir.Loop)]
    assert len(loops) == 1
    assert sorted([str(v) for v in loops[0][1]]) == ['i', 'x']

    # Verify that the variables in the statements are found
    stmts = [v for v in variables if isinstance(v[0], ir.Assignment)]
    assert len(stmts) == 2

    assert sorted([str(v) for v in stmts[0][1]]) == ['i', 'i', 'scalar', 'vector(i)', 'vector(i)']
    assert sorted([str(v) for v in stmts[1][1]]) == ['i', 'i', 'i', 'matrix(i, :)', 'vector(i)']


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_finder_unique_with_ir_node(frontend):
    """
    Test the expression finder's ability to yield the ir node combined with only unique
    variables.
    """
    fcode = """
subroutine routine_simple (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i

  do i=1, x
     vector(i) = vector(i) + scalar
     matrix(i, :) = i * vector(i)
  end do
end subroutine routine_simple
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    variables = FindVariables(with_ir_node=True).visit(routine.body)
    assert len(variables) == 3
    assert all(isinstance(v, tuple) and len(v) == 2 for v in variables)

    # Verify that the variables in the loop definition are found
    loops = [v for v in variables if isinstance(v[0], ir.Loop)]
    assert len(loops) == 1
    assert sorted([str(v) for v in loops[0][1]]) == ['i', 'x']

    # Verify that the variables in the statements are found
    stmts = [v for v in variables if isinstance(v[0], ir.Assignment)]
    assert len(stmts) == 2

    assert sorted([str(v) for v in stmts[0][1]]) == ['i', 'scalar', 'vector(i)']
    assert sorted([str(v) for v in stmts[1][1]]) == ['i', 'matrix(i, :)', 'vector(i)']


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_callback_mapper(frontend):
    """
    Test the ExpressionFinder together with ExpressionCallbackMapper. This is just a very basic
    sanity check and does not cover all angles.
    """
    fcode = """
subroutine routine_simple (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i, j

  do i=1, x
    vector(i) = vector(i) + scalar
    do j=1, y
      if (j > i) then
        matrix(i, j) = real(i * j, kind=jprb) + 1.
      else
        matrix(i, j) = i * vector(j)
      end if
    end do
  end do
end subroutine routine_simple
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    # Nonsense example that singles out anything that is a matrix
    def is_matrix(expr, *args, **kwargs):  # pylint: disable=unused-argument
        if isinstance(expr, sym.Array) and expr.type.shape and len(expr.type.shape) == 2:
            return expr
        return None

    class FindMatrix(ExpressionFinder):
        retriever = ExpressionCallbackMapper(
            callback=is_matrix,
            combine=lambda v: tuple(e for e in v if e is not None)
        )

    matrix_count = FindMatrix(unique=False).visit(routine.body)
    assert len(matrix_count) == 2

    matrix_count = FindMatrix().visit(routine.body)
    assert len(matrix_count) == 1
    assert str(matrix_count.pop()) == 'matrix(i, j)'


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_retriever_recurse_query(frontend):
    """
    Test the ExpressionRetriever with a custom recurse query that allows to terminate recursion
    early.
    """
    fcode = """
subroutine routine_simple (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i, j

  do i=1, x
    vector(i) = vector(i) + scalar
    do j=1, y
      if (j > i) then
        matrix(i, j) = real(i * j + 2, kind=jprb) + 1.
      else
        matrix(i, j) = i * vector(j)
      end if
    end do
  end do
end subroutine routine_simple
"""

    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    # Find all literals except when they appear in array subscripts or loop ranges
    class FindLiteralsNotInSubscriptsOrRanges(ExpressionFinder):
        retriever = ExpressionRetriever(
            query=lambda expr: isinstance(expr, (sym.IntLiteral, sym.FloatLiteral, sym.LogicLiteral)),
            recurse_query=lambda expr, *args, **kwargs: not isinstance(expr, (sym.ArraySubscript, sym.LoopRange))
        )
    literals = FindLiteralsNotInSubscriptsOrRanges(unique=False).visit(routine.body)

    if frontend == OMNI:
        # OMNI substitutes jprb
        assert len(literals) == 4
        assert sorted([str(l) for l in literals]) == ['1.', '13', '2', '300']
    else:
        assert len(literals) == 2
        assert sorted([str(l) for l in literals]) == ['1.', '2']


@pytest.mark.parametrize('frontend', available_frontends())
def test_find_variables_associates(frontend):
    """
    Test correct discovery of variables in associates.
    """
    fcode = """
subroutine find_variables_associates (x, y, scalar, vector, matrix)
  integer, parameter :: jprb = selected_real_kind(13,300)
  integer, intent(in) :: x, y
  real(kind=jprb), intent(in) :: scalar
  real(kind=jprb), intent(inout) :: vector(x), matrix(x, y)
  integer :: i, j

  do i=1, x
    associate (v => vector(i), m => matrix(i, :))
      vector(i) = vector(i) + scalar
      do j=1, y
        if (j > i) then
          m(j) = real(i * j, kind=jprb) + 1.
        else
          matrix(i, j) = i * vector(j)
        end if
      end do
    end associate
  end do
end subroutine find_variables_associates
"""
    # Test the internals of the subroutine
    routine = Subroutine.from_source(fcode, frontend=frontend)

    variables = FindVariables(unique=False).visit(routine.body)
    assert len(variables) == 27 if frontend == OMNI else 28
    assert len([v for v in variables if v.name == 'v']) == 1
    assert len([v for v in variables if v.name == 'm']) == 2


@pytest.mark.parametrize('frontend', available_frontends())
def test_is_parent_of(frontend):
    """
    Test the ``is_parent_of`` utility.
    """
    fcode = """
subroutine test_is_parent_of
  implicit none
  integer :: a, j, n=10

  a = 0
  do j=1,n
    if (j > 3) then
      a = a + 1
    end if
  end do
end subroutine test_is_parent_of
    """.strip()
    routine = Subroutine.from_source(fcode, frontend=frontend)

    loop = FindNodes(ir.Loop).visit(routine.body)[0]
    conditional = FindNodes(ir.Conditional).visit(routine.body)[0]
    assignments = FindNodes(ir.Assignment).visit(routine.body)

    assert is_parent_of(loop, conditional)
    assert not is_parent_of(conditional, loop)

    for node in [loop, conditional]:
        assert {is_parent_of(node, a) for a in assignments} == {True, False}
        assert all(not is_parent_of(a, node) for a in assignments)


@pytest.mark.parametrize('frontend', available_frontends())
def test_is_child_of(frontend):
    """
    Test the ``is_child_of`` utility.
    """
    fcode = """
subroutine test_is_child_of
  implicit none
  integer :: a, j, n=10

  a = 0
  do j=1,n
    if (j > 3) then
      a = a + 1
    end if
  end do
end subroutine test_is_child_of
    """.strip()
    routine = Subroutine.from_source(fcode, frontend=frontend)

    loop = FindNodes(ir.Loop).visit(routine.body)[0]
    conditional = FindNodes(ir.Conditional).visit(routine.body)[0]
    assignments = FindNodes(ir.Assignment).visit(routine.body)

    assert not is_child_of(loop, conditional)
    assert is_child_of(conditional, loop)

    for node in [loop, conditional]:
        assert {is_child_of(a, node) for a in assignments} == {True, False}
        assert all(not is_child_of(node, a) for a in assignments)


@pytest.mark.parametrize('frontend', available_frontends())
def test_attach_scopes_associates(frontend, tmp_path):
    fcode = """
module attach_scopes_associates_mod
    implicit none

    type other_type
        integer :: foo
    end type other_type

    type some_type
        type(other_type) :: var
    end type some_type

contains

    subroutine attach_scopes_associates
        type(some_type) :: blah
        integer :: a

        associate(var=>blah%var)
            associate(bar=>5+3)
                a = var%foo
            end associate
        end associate
    end subroutine attach_scopes_associates
end module attach_scopes_associates_mod
    """.strip()

    module = Module.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    routine = module['attach_scopes_associates']
    associates = FindNodes(ir.Associate).visit(routine.body)
    assert len(associates) == 2
    assignment = FindNodes(ir.Assignment).visit(routine.body)
    assert len(assignment) == 1
    assert len(FindVariables().visit(assignment)) == 3
    var_map = {str(var): var for var in FindVariables().visit(assignment)}
    assert len(var_map) == 3
    assert associates[1].parent is associates[0]
    assert var_map['a'].scope is routine
    assert var_map['var%foo'].scope is associates[0]
    assert var_map['var%foo'].parent.scope is associates[0]
    assert var_map['var%foo'].parent is var_map['var']



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Collection of utilities to automatically remove code elements or
section and to perform Dead Code Elimination.
"""

import operator as op
from itertools import chain

from loki.analyse import dataflow_analysis_attached
from loki.batch import Transformation
from loki.expression import simplify, symbols as sym, symbolic_op
from loki.ir import nodes as ir, Transformer, FindNodes, FindVariables, FindInlineCalls, SubstituteExpressions
from loki.ir.pragma_utils import (
    is_loki_pragma, pragma_regions_attached, get_pragma_parameters
)
from loki.program_unit import ProgramUnit
from loki.tools import flatten, as_tuple, OrderedSet
__all__ = [
    'RemoveCodeTransformation',
    'do_remove_dead_code', 'RemoveDeadCodeTransformer',
    'do_remove_marked_regions', 'RemoveRegionTransformer',
    'do_remove_calls', 'RemoveCallsTransformer', 'do_remove_unused_vars'
]


class RemoveCodeTransformation(Transformation):
    """
    A :any:`Transformation` that provides named call and import
    removal, code removal of pragma-marked regions and Dead Code
    Elimination for batch processing via the :any:`Scheduler`.

    The transformation will apply the following methods in order:

    * :any:`do_remove_calls`
    * :any:`do_remove_marked_regions`
    * :any:`do_remove_dead_code`
    * :any:`do_remove_unused_vars`

    Parameters
    ----------
    remove_marked_regions : boolean
        Flag to trigger the use of :meth:`remove_marked_regions`;
        default: ``True``
    mark_with_comment : boolean
        Flag to trigger the insertion of a marker comment when
        removing a region; default: ``True``.
    replacement_call : optional, str
        Name of the "abort" subroutine to call if a replacement call
        is to be inserted in :meth:`do_remove_marked_regions`.
    replacement_msg : optional, str
        Optional error message that will be passed as argument to
        the replacement call in :meth:`do_remove_marked_regions`.
    replacement_module : optional, str
        Optional name of the module from which to import the
        replacement subroutine in :meth:`do_remove_marked_regions`.
    remove_dead_code : boolean
        Flag to trigger the use of :meth:`remove_dead_code`;
        default: ``False``
    use_simplify : boolean
        Use :any:`simplify` when branch pruning in during
        :meth:`remove_dead_code`.
    call_names : list of str
        List of subroutine names against which to match
        :any:`CallStatement` nodes during :meth:`remove_calls`.
    intrinsic_names : list of str
        List of module names against which to match :any:`GenericStmt`
        nodes during :meth:`remove_calls`.
    remove_imports : boolean
        Flag indicating whether to remove symbols from :any:`Import`
        objects during :meth:`remove_calls`; default: ``True``
    kernel_only : boolean
        Only apply the configured removal to subroutines marked as
        "kernel"; default: ``False``
    remove_unused_args : boolean
        Remove unused dummy arguments from routines.
    remove_unused_vars : boolean
        Remove unused variables/locals from routines.
    remove_only_arrays : boolean
        Whether to only remove unused arrays from routines
        or all variables/locals.
    """

    _key = 'RemoveCodeTransformation'

    # Recurse to subroutines in ``contains`` clause
    recurse_to_internal_procedures = True
    reverse_traversal = True

    def __init__(
            self, remove_marked_regions=True, mark_with_comment=True,
            replacement_call=None, replacement_msg=None, replacement_module=None,
            remove_dead_code=False, use_simplify=True,
            call_names=None, intrinsic_names=None,
            remove_imports=True, kernel_only=False,
            remove_unused_args=False, remove_unused_vars=False,
            remove_only_arrays=True
    ):
        self.remove_marked_regions = remove_marked_regions
        self.mark_with_comment = mark_with_comment
        self.replacement_call = replacement_call
        self.replacement_msg = replacement_msg
        self.replacement_module = replacement_module

        self.remove_dead_code = remove_dead_code
        self.use_simplify = use_simplify

        self.call_names = as_tuple(call_names)
        self.intrinsic_names = as_tuple(intrinsic_names)
        self.remove_imports = remove_imports

        self.kernel_only = kernel_only
        self.remove_unused_args = remove_unused_args

        self.remove_unused_vars = remove_unused_vars
        self.remove_only_arrays = remove_only_arrays

    def transform_subroutine(self, routine, **kwargs):

        if kwargs.get('role') == 'kernel' or not self.kernel_only:
            # Apply named node removal to strip specific calls
            if self.call_names or self.intrinsic_names:
                do_remove_calls(
                    routine, call_names=self.call_names,
                    intrinsic_names=self.intrinsic_names,
                    remove_imports=self.remove_imports
                )

            # Apply marked region removal
            if self.remove_marked_regions:
                do_remove_marked_regions(
                    routine, mark_with_comment=self.mark_with_comment,
                    replacement_call=self.replacement_call,
                    replacement_msg=self.replacement_msg,
                    replacement_module=self.replacement_module
                )

            # Apply Dead Code Elimination
            if self.remove_dead_code:
                do_remove_dead_code(routine, use_simplify=self.use_simplify)

            if self.remove_unused_vars:
                do_remove_unused_vars(routine, remove_only_arrays=self.remove_only_arrays)

        if self.remove_unused_args and (item := kwargs['item']):
            # collect unused args from successors
            successors = kwargs['sub_sgraph'].successors(item=item)
            unused_args_map = {}
            for successor in successors:
                succ_data = successor.trafo_data.get(self._key, {})
                succ_ir = successor.ir
                unused_args_map[succ_ir] = succ_data.get(succ_ir.name, {}).get('unused_args', {})

            variable_map = routine.symbol_map
            used_or_defined_symbols = OrderedSet()
            if item.config.get('remove_unused_args', True) and kwargs['role'] == 'kernel':
                # internal procedures are transformed after the parent routine,
                # so determine their unused args eagerly to update call sites here
                for child in routine.subroutines:
                    child_symbol_map = child.symbol_map
                    child_used_symbols = get_used_or_defined_symbols(child)

                    # Collect symbols used or defined in contained subroutines, whilst ignoring
                    # those that mask parent symbols
                    used_or_defined_symbols |= OrderedSet(
                        variable_map[child_symbol.name_parts[0]].clone(dimensions=None)
                        for child_symbol in child_used_symbols
                        if child_symbol.name_parts[0] in variable_map
                        and child_symbol.name_parts[0] not in child_symbol_map
                    )

                    child_unused_args, _ = find_unused_dummy_args_and_vars(
                        child, used_or_defined_symbols=child_used_symbols
                    )
                    unused_args_map[child] = child_unused_args
                    item.trafo_data.setdefault(self._key, {})[child.name] = {'unused_args': child_unused_args}

            do_remove_unused_call_args(routine, unused_args_map)

            if item.config.get('remove_unused_args', True) and kwargs['role'] == 'kernel':
                # find unused args, which for contained subroutines may already be known
                item_data = item.trafo_data.get(self._key, {})
                unused_args = item_data.get(routine.name, {}).get('unused_args', {})

                if not unused_args:
                    used_or_defined_symbols |= get_used_or_defined_symbols(routine)
                    unused_args, _ = find_unused_dummy_args_and_vars(
                        routine, used_or_defined_symbols=used_or_defined_symbols
                    )

                    # store unused args keyed by routine name to avoid
                    # internal procedures overwriting the parent's data
                    item.trafo_data.setdefault(self._key, {})[routine.name] = {'unused_args': unused_args}

                do_remove_unused_dummy_args(routine, unused_args)


def do_remove_unused_dummy_args(routine, unused_args):
    """
    Utility routine to remove unused dummy arguments from
    a given routine.

    Parameters
    ----------
    routine : :any:`Subroutine`
        A :any:`Subroutine` whose unused dummy arguments will be removed.
    unused_args : dict
       A dict mapping the unused dummy argument symbol to its position in the
       routine's argument list.This must be retrieved using the
       :any:`find_unused_dummy_args_and_vars` utility.
    """

    routine.variables = [a for a in routine.variables
                         if not a.name.lower() in unused_args]

def do_remove_unused_vars(routine, unused_vars=None, remove_only_arrays=True):
    """
    Utility routine to remove unused variables (or only local arrays) from a given routine.

    Parameters
    ----------
    routine : :any:`Subroutine`
        A :any:`Subroutine` whose unused dummy arguments will be removed.
    unused_args : dict, optional
        A list of unused vars. This can be retrieved using the
       :any:`find_unused_dummy_args_and_vars` utility.
    remove_only_arrays : bool, optional
        Whether to only remove arrays or all variables/temporaries
        that are unused within the routine.
    """
    if unused_vars is None:
        _, unused_vars = find_unused_dummy_args_and_vars(routine)
    if remove_only_arrays:
        unused_vars = [var for var in unused_vars if isinstance(var, sym.Array)]
    routine.variables = [var for var in routine.variables
                         if not var.name.lower() in unused_vars]

def do_remove_unused_call_args(routine, unused_args_map):
    """
    Utility routine to remove unused arguments from all the
    :any:`CallStatement`s and :any:`InlineCall`s in a given routine.

    Parameters
    ----------
    routine : :any:`Subroutine`
        A :any:`Subroutine` whose call statements and inline calls will be updated.
    unused_args_map : dict
       A dict mapping the :any:`Subroutine` corresponding to the :any:`CallStatement`
       or :any:`InlineCall`, accessed via the ``routine`` property, to its unused arguments.
       The unused arguments must be retrieved using the :any:`find_unused_dummy_args`
       utility.
    """

    inline_call_map = {}
    for call in chain(FindNodes(ir.CallStatement).visit(routine.body), FindInlineCalls().visit(routine.body)):
        if not call.routine or not unused_args_map.get(call.routine, None):
            continue

        unused_positions = {c for c in unused_args_map[call.routine].values() if c < len(call.arguments)}
        unused_kwargs = [(kw, arg) for kw, arg in call.kwarguments if kw.lower() in unused_args_map[call.routine]]

        new_args = [arg for i, arg in enumerate(call.arguments) if i not in unused_positions]
        new_kwargs = [(kw, arg) for kw, arg in call.kwarguments if not (kw, arg) in unused_kwargs]

        if isinstance(call, ir.CallStatement):
            call._update(arguments=as_tuple(new_args), kwarguments=as_tuple(new_kwargs))
        else:
            inline_call_map[call] = call.clone(parameters=as_tuple(new_args), kw_parameters=dict(new_kwargs))

    if inline_call_map:
        routine.body = SubstituteExpressions(inline_call_map).visit(routine.body)


def get_used_or_defined_symbols(routine):
    """
    Collect symbols used or defined in a routine, including array-shape references.
    """
    variable_map = routine.symbol_map
    routine_arg_names = routine.argnames
    with dataflow_analysis_attached(routine):
        used_or_defined_symbols = routine.body.uses_symbols | routine.body.defines_symbols

        # We search for symbols used to define array sizes of symbols referenced
        # in the body, as well as local arrays declared in the routine.
        used_or_defined_array_shapes = [s.shape for s in used_or_defined_symbols if isinstance(s, sym.Array)]
        local_var_shapes = [
            var.shape for var in routine.variables
            if isinstance(var, sym.Array) and var.name.lower() not in routine_arg_names
        ]
        used_or_defined_symbols |= FindVariables(unique=True).visit(
            used_or_defined_array_shapes + local_var_shapes
        )

        used_or_defined_symbols |= OrderedSet(
            variable_map.get(v.name_parts[0], v).clone(dimensions=None)
            for v in used_or_defined_symbols
        )

    return used_or_defined_symbols


def find_unused_dummy_args_and_vars(routine, used_or_defined_symbols=None):
    """
    Utility routine to find all the unused arguments in a :any:`Subroutine`.

    Parameters
    ----------
    routine : :any:`Subroutine`
        A :any:`Subroutine` to search for unused dummy arguments.
    used_or_defined_symbols : set, optional
        The set of symbols used or defined in the routine.

    Return
    ------
    unused_args : dict
       A dict mapping the unused dummy argument symbol to its position in the
       routine's argument list.
    """

    if used_or_defined_symbols is None:
        used_or_defined_symbols = get_used_or_defined_symbols(routine)

    unused_args = {a.clone(dimensions=None): c for c, a in enumerate(routine.arguments)
                   if not a.name.lower() in used_or_defined_symbols}
    routine_arg_names = routine.argnames
    local_vars = [var for var in routine.variables if var.name.lower() not in routine_arg_names]
    unused_vars = [var.clone(dimensions=None) for var in local_vars
            if not var.name.lower() in used_or_defined_symbols]

    return unused_args, unused_vars


def do_remove_dead_code(routine, use_simplify=True):
    """
    Perform Dead Code Elimination on the given :any:`Subroutine` object.

    Parameters
    ----------
    routine : :any:`Subroutine`
        The subroutine to which to apply dead code elimination.
    simplify : boolean
        Use :any:`simplify` when evaluating expressions for branch pruning.
    """

    transformer = RemoveDeadCodeTransformer(use_simplify=use_simplify)
    routine.body = transformer.visit(routine.body)


class RemoveDeadCodeTransformer(Transformer):
    """
    :any:`Transformer` class that removes provably unreachable code paths.

    The primary modification performed is to prune individual code branches
    under :any:`Conditional` nodes.

    Parameters
    ----------
    use_simplify : boolean
        Use :any:`simplify` when evaluating expressions for branch pruning.
    """

    def __init__(self, use_simplify=True, **kwargs):
        super().__init__(**kwargs)
        self.use_simplify = use_simplify

    def visit_Conditional(self, o, **kwargs):
        condition = self.visit(o.condition, **kwargs)
        body = as_tuple(flatten(as_tuple(self.visit(o.body, **kwargs))))
        else_body = as_tuple(flatten(as_tuple(self.visit(o.else_body, **kwargs))))

        if self.use_simplify:
            condition = simplify(condition)

        if condition == 'True':
            return body

        if condition == 'False':
            return else_body

        has_elseif = o.has_elseif and else_body and isinstance(else_body[0], ir.Conditional)
        return self._rebuild(o, tuple((condition,) + (body,) + (else_body,)), has_elseif=has_elseif)

    def visit_MultiConditional(self, o, **kwargs):
        # Get select expression and simplify if requested
        expr = self.visit(o.expr, **kwargs)
        expr = simplify(expr) if self.use_simplify else expr

        values = self.visit(o.values, **kwargs)
        bodies = self.visit(o.bodies, **kwargs)
        else_body = self.visit(o.else_body, **kwargs)

        for val, body in zip(values, bodies):
            # Equate select expression against case values
            for v in val:
                if symbolic_op(expr, op.eq, v):
                    return body

        if expr == 'False':
            # Simplify to default if always false
            return else_body

        return self._rebuild(o, tuple((expr,) + (values,) + (bodies,) + (else_body,)), name=o.name)


def do_remove_marked_regions(
        routine, mark_with_comment=True, replacement_call=None,
        replacement_msg=None, replacement_module=None
):
    """
    Utility routine to remove code regions marked with
    ``!$loki remove`` pragmas from a subroutine's body.

    Optionally, any removed region might be marked with a
    comment and/or a simple single-argument "abort" call. For this,
    a subroutine name and message can be specified and an optional
    check for an import can also be defined to ensure the interface
    for the abort procedure is available. To bypass the replacement
    call insertion for individual pragma regions use
    ``!$loki remove no-replacement-call``.

    Parameters
    ----------
    routine : :any:`Subroutine`
        The subroutine to which to apply dead code elimination.
    mark_with_comment : boolean
        Flag to trigger the insertion of a marker comment when
        removing a region; default: ``True``.
    replacement_call : optional, str
        Name of the "abort" subroutine to call if a replacement call
        is to be inserted.
    replacement_msg : optional, str
        Optional error message that will be passed as a single
        argument to the replacement call.
    replacement_module : optional, str
        Optional name of the module from which to import the
        replacement subroutine. This will only be inserted if a
        replacement was perfored and will not replace existing imports
        of the same module or symbols.
    """

    transformer = RemoveRegionTransformer(
        mark_with_comment=mark_with_comment,
        replacement_call=replacement_call,
        replacement_msg=replacement_msg,
    )

    with pragma_regions_attached(routine, keyword='loki'):
        routine.body = transformer.visit(routine.body, scope=routine)

    if transformer.replacement_done and replacement_module:
        # Get newly inject procedure symbol for the replacement call
        callsym = sym.ProcedureSymbol(replacement_call, scope=routine)

        # Inject import of replacement module if it does not exist
        import_map = {i.module: i for i in routine.imports}
        if imprt := import_map.get(replacement_module):
            if not any(s == replacement_call for s in imprt.symbols):
                imprt._update(symbols=imprt.symbols + (callsym,))
        else:
            routine.spec.prepend(ir.Import(
                module=f'{replacement_module}', symbols=(callsym,), c_import=False)
            )


class RemoveRegionTransformer(Transformer):
    """
    A :any:`Transformer` that removes code regions marked with
    ``!$loki remove`` pragmas.

    This :any:`Transformer` only removes :any:`PragmaRegion` nodes,
    and thus requires the IR tree to have pragma regions attached, for
    example via :meth:`pragma_regions_attached`.

    When removing a marked code region the transformer may leave a
    comment or a replacement call to trigger "abort" errors in the
    source to mark the previous location.

    Parameters
    ----------
    mark_with_comment : boolean
        Flag to trigger the insertion of a marker comment when
        removing a region; default: ``True``.
    replacement_call : optional, str
        Name of the "abort" subroutine to call if a replacement call
        is to be inserted.
    replacement_msg : optional, str
        Optional error message that will be passed as a single
        argument to the replacmeent call.
    """

    def __init__(
            self, mark_with_comment=True, replacement_call=None, replacement_msg=None, **kwargs
    ):
        super().__init__(**kwargs)

        self.mark_with_comment = mark_with_comment

        # Replace section with call to trigger abort messages!
        self.replacement_call = replacement_call
        self.replacement_msg = replacement_msg
        self.replacement_done = False

    def visit_PragmaRegion(self, o, **kwargs):
        """ Remove :any:`PragmaRegion` nodes with ``!$loki remove`` pragmas """

        # Skip if the bypass clause is present
        bypass = 'no-replacement-call' in get_pragma_parameters(o.pragma, starts_with='remove')
        if is_loki_pragma(o.pragma, starts_with='remove'):

            # Leave a comment to mark the removed region in source
            replacement = []
            if self.mark_with_comment:
                replacement.append(ir.Comment(text='! [Loki] Removed content of pragma-marked region!'))

            if self.replacement_call and not bypass:
                # Get the outer scope, to avoid picking associates
                routine = kwargs['scope']
                while not isinstance(routine, ProgramUnit):
                    routine = routine.parent

                # If requested add a call to a simple subroutine with an error message arg
                replacement.append(ir.CallStatement(
                    name=sym.ProcedureSymbol(self.replacement_call, scope=routine),
                    arguments=sym.Literal(str(self.replacement_msg.format(routine.name)))
                ))
                # Set a flag to trigger import injections
                self.replacement_done = True

            return as_tuple(replacement)

        # Recurse into the pragama region and rebuild
        rebuilt = tuple(self.visit(i, **kwargs) for i in o.children)
        return self._rebuild(o, rebuilt)


def do_remove_calls(
        routine, call_names=None, intrinsic_names=None, remove_imports=True
):
    """
    Utility routine to remove all :any:`CallStatement` nodes
    to specific named subroutines in a :any:`Subroutine`.

    For more information, see :any:`RemoveCallsTransformer`.

    Parameters
    ----------
    call_names : list of str
        List of subroutine names against which to match
        :any:`CallStatement` nodes.
    intrinsic_names : list of str
        List of module names against which to match :any:`GenericStmt`
        nodes.
    remove_imports : boolean
        Flag indicating whether to remove the respective procedure
        symbols from :any:`Import` objects; default: ``True``.
    """

    transformer = RemoveCallsTransformer(
        call_names=call_names, intrinsic_names=intrinsic_names,
        remove_imports=remove_imports
    )
    routine.spec = transformer.visit(routine.spec)
    routine.body = transformer.visit(routine.body)


class RemoveCallsTransformer(Transformer):
    """
    A :any:`Transformer` that removes all :any:`CallStatement` nodes
    to specific named subroutines.

    This :any:`Transformer` will by default also remove the enclosing
    inline-conditional when encountering calls of the form ```if
    (flag) call named_procedure()``.

    This :any:`Transformer` will also attempt to match and remove
    :any:`GenericStmt` nodes against a given list of name strings.  This
    allows removing intrinsic calls like ``write (*,*) "..."``.

    In addition, this :any:`Transformer` can also attempt to match and
    remove :any:`Import` nodes if given a list of strings to
    match. This can be used to remove the associated imports of the
    removed subroutines.

    Parameters
    ----------
    call_names : list of str
        List of subroutine names against which to match
        :any:`CallStatement` nodes.
    intrinsic_names : list of str
        List of module names against which to match :any:`GenericStmt`
        nodes.
    remove_imports : boolean
        Flag indicating whether to remove the respective procedure
        symbols from :any:`Import` objects; default: ``True``.
    """

    def __init__(
            self, call_names=None, intrinsic_names=None,
            remove_imports=True, **kwargs
    ):
        super().__init__(**kwargs)

        self.call_names = as_tuple(call_names)
        self.intrinsic_names = as_tuple(intrinsic_names)
        self.remove_imports = remove_imports

    def visit_CallStatement(self, o, **kwargs):
        """ Match and remove :any:`CallStatement` nodes against name patterns """
        if o.name in self.call_names:
            return None

        rebuilt = tuple(self.visit(i, **kwargs) for i in o.children)
        return self._rebuild(o, rebuilt)

    def visit_Conditional(self, o, **kwargs):
        """ Remove inline-conditionals after recursing into their body """

        # First, recurse into condition and bodies
        cond, body, else_body = tuple(self.visit(i, **kwargs) for i in o.children)

        # Capture and remove newly empty inline conditionals
        if o.inline and len(body) == 0:
            return None

        return self._rebuild(o, (cond, body, else_body))

    def visit_GenericStmt(self, o, **kwargs):
        """ Match and remove :any:`GenericStmt` nodes against name patterns """
        if self.intrinsic_names:
            if any(str(c).lower() in str(o.text).lower() for c in self.intrinsic_names):
                return None

        rebuilt = tuple(self.visit(i, **kwargs) for i in o.children)
        return self._rebuild(o, rebuilt)

    def visit_Import(self, o, **kwargs):
        """ Remove the symbol of any named calls from Import nodes """

        symbols_found = any(s in self.call_names for s in o.symbols)
        if self.remove_imports and symbols_found:
            new_symbols = tuple(s for s in o.symbols if s not in self.call_names)
            return o.clone(symbols=new_symbols) if new_symbols else None

        rebuilt = tuple(self.visit(i, **kwargs) for i in o.children)
        return self._rebuild(o, rebuilt)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from loki.backend.fgen import FortranCodegen
from loki.frontend.source import SourceStatus
from loki.tools.util import as_tuple


__all__ = ['FortranCodegenConservative']


class FortranCodegenConservative(FortranCodegen):
    """
    Strictly conservative version of :any:`FortranCodegen` visitor
    that will attempt to use existing :any:`Source` information from
    the frontends where possible.
    """

    def visit_Node(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string
        return super().visit_Node(o, *args, **kwargs)

    def visit_Assignment(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string

        if o.source and o.source.status == SourceStatus.INVALID_NODE:
            # Attempt to recreate some structure by locating `=`
            slist = o.source.string.split('=', maxsplit=1)
            assert len(slist) == 2

            # If LHS hasn't changed, prefer source
            lhs = self.visit(o.lhs, **kwargs) + ' '
            if slist[0] == o.lhs:
                lhs = slist[0]
            else:
                # At least prescribe previous indent...
                ind = len(o.source.string) - len(o.source.string.lstrip())
                lhs = ind*' ' + lhs

            # If RHS hasn't changed, prefer source
            rhs = ' ' + self.visit(o.rhs, **kwargs)
            if slist[1] == o.rhs:
                rhs = slist[1]

            comment = str(self.visit(o.comment, **kwargs)) if o.comment else ''
            if o.ptr:
                return self.format_line(lhs, '=>', rhs, comment=comment)

            return self.format_line(lhs, '=', rhs, comment=comment, no_indent=True)

        return super().visit_Assignment(o, *args, **kwargs)

    def visit_CallStatement(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string
        return super().visit_CallStatement(o, *args, **kwargs)

    def visit_Comment(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            slist = o.source.string.split('!', maxsplit=1)
            pre, txt = (slist[0], slist[1]) if len(slist) > 1 else (None, slist[0])
            if pre and not pre.isspace():
                # An inline comment, only return text and leading whitespace
                ws = ' '*(len(pre)-len(pre.rstrip(' ')))
                return f'{ws}!{txt}'
            return o.source.string
        return super().visit_Comment(o, *args, **kwargs)

    def visit_Conditional(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string

        if o.source and o.source.status == SourceStatus.INVALID_CHILDREN:
            if o.inline:
                # TODO: Deal with inline conditionals properly
                return super().visit_Conditional(o, *args, **kwargs)

            header = o.source.string.splitlines()[0]

            self.depth += self.style.conditional_indent
            body = self.visit(o.body, **kwargs)
            if o.has_elseif:
                self.depth -= self.style.conditional_indent
                else_body = [self.visit(o.else_body, is_elseif=True, **kwargs)]
            else:
                else_body = [self.visit(o.else_body, **kwargs)]
                self.depth -= self.style.conditional_indent
                if o.else_body:
                    # Get the `ELSE` from source to get its indentation
                    elseline = [
                        s for s in o.source.string.splitlines()
                        if s.upper().strip() == 'ELSE'
                    ]
                    else_body = [elseline[-1]] + else_body

                # Recapture the footer line from source
                footer = o.source.string.splitlines()[o.source.lines[1]-o.source.lines[0]]
                else_body += [footer]

            return self.join_lines(header, body, *else_body)

        return super().visit_Conditional(o, *args, **kwargs)

    def visit_VariableDeclaration(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string

        if o.source and o.source.status == SourceStatus.INVALID_NODE:
            no_indent = False

            # Attempt to recreate some structure by locating `::`
            slist = o.source.string.split('::', maxsplit=1)
            assert len(slist) == 2

            # If type attributes haven't changed, prefer source (and don't indent)
            attributes = str(self.join_items(self._construct_type_attributes(o, **kwargs))) + ' '
            # TODO: Type attributes don't string compare cleanly yet; needs fixing!
            if slist[0].strip().lower() == attributes.strip().lower():
                attributes = slist[0]
                no_indent = True

            # If declared variables haven't changed, prefer source
            variables = ' ' + str(self.join_items(self._construct_decl_variables(o, **kwargs)))
            if o.symbols == as_tuple(slist[1].split(',')):
                variables = slist[1]

            comment = str(self.visit(o.comment, **kwargs)) if o.comment else ''

            return self.format_line(attributes, '::', variables, comment, no_indent=no_indent)

        return super().visit_VariableDeclaration(o, *args, **kwargs)

    def visit_Import(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string
        return super().visit_Import(o, *args, **kwargs)

    def visit_Loop(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string

        if o.source and o.source.status == SourceStatus.INVALID_CHILDREN:
            # Recapture header and footer from source
            header = o.source.string.splitlines()[0]
            footer = o.source.string.splitlines()[o.source.lines[1]-o.source.lines[0]]

            pragma = self.visit(o.pragma, **kwargs)
            pragma_post = self.visit(o.pragma_post, **kwargs)
            self.depth += self.style.loop_indent
            body = self.visit(o.body, **kwargs)
            self.depth -= self.style.loop_indent
            return self.join_lines(pragma, header, body, footer, pragma_post)

        return super().visit_Loop(o, *args, **kwargs)

    def visit_Section(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string
        return super().visit_Section(o, *args, **kwargs)

    def visit_Subroutine(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string

        if o.source and o.source.status == SourceStatus.INVALID_CHILDREN:
            # Re-construct header and footer from source if possible
            h_end = o.body.source.lines[0] if o.body.source else o.source.lines[1]
            h_end = min(h_end, o.spec.source.lines[0]) if o.spec.source else h_end
            if o.docstring:
                h_end = min(h_end, o.docstring[0].source.lines[0])

            if h_end < o.source.lines[1]:
                header = '\n'.join(o.source.string.splitlines()[:h_end-o.source.lines[0]])
            else:
                header = self._construct_subroutine_header(o, **kwargs)

            # For one-line footers reconstruct from source
            foot = o.source.string.splitlines()[o.source.lines[1]-o.source.lines[0]]
            if 'END ' in foot.upper():
                footer = foot
            else:
                footer = self._construct_procedure_footer(o, **kwargs)

            self.depth += self.style.procedure_spec_indent
            docstring = self.visit(o.docstring, **kwargs)
            spec = self.visit(o.spec, **kwargs)
            self.depth -= self.style.procedure_spec_indent

            self.depth += self.style.procedure_body_indent
            body = self.visit(o.body, **kwargs)
            self.depth -= self.style.procedure_body_indent

            self.depth += self.style.procedure_contains_indent
            contains = self.visit(o.contains, **kwargs)
            self.depth -= self.style.procedure_contains_indent
            if contains:
                return self.join_lines(header, docstring, spec, body, contains, footer)

            return self.join_lines(header, docstring, spec, body, footer)

        return super().visit_Subroutine(o, *args, **kwargs)

    def visit_Module(self, o, *args, **kwargs):
        if o.source and o.source.status == SourceStatus.VALID:
            return o.source.string

        if o.source and o.source.status == SourceStatus.INVALID_CHILDREN:
            # Re-construct header and footer from source if possible
            h_end = o.contains.source.lines[0] if o.contains and o.contains.source else o.source.lines[1]
            h_end = min(h_end, o.spec.source.lines[0]) if o.spec.source else h_end

            if h_end < o.source.lines[1]:
                header = '\n'.join(o.source.string.splitlines()[:h_end-o.source.lines[0]])
            else:
                header = self._construct_module_header(o, **kwargs)

            # For one-line footers reconstruct from source
            foot = o.source.string.splitlines()[o.source.lines[1]-o.source.lines[0]]
            if 'END ' in foot.upper():
                footer = foot
            else:
                footer = self._construct_module_footer(o, **kwargs)

            self.depth += self.style.module_spec_indent
            spec = self.visit(o.spec, **kwargs)
            self.depth -= self.style.module_spec_indent

            self.depth += self.style.module_contains_indent
            contains = self.visit(o.contains, **kwargs)
            self.depth -= self.style.module_contains_indent
            if contains:
                return self.join_lines(header, spec, contains, footer)

            return self.join_lines(header, spec, footer)

        return super().visit_Module(o, *args, **kwargs)
